"""
Dashboard definition classes.

This module provides several classes to define a Dashboard.
"""

from operator import attrgetter
from powa.compat import classproperty, with_metaclass
from powa.framework import AuthHandler
from powa.json import JSONizable
from powa.ui_modules import MenuEntry
from tornado.web import URLSpec

try:
    from collections import OrderedDict
except ModuleNotFoundError:
    from ordereddict import OrderedDict

import psycopg2
from inspect import isfunction

GLOBAL_COUNTER = 0


class DashboardHandler(AuthHandler):
    """
    Handler for a specific dashboard
    """

    def initialize(self, template, params):
        self.template = template
        self.params = params

    def get(self, *args):
        # Compatibility for tornado < 3:
        if not hasattr(self, "path_args"):
            self.path_args = args

        params = OrderedDict(zip(self.params, args))
        title = self.dashboard().title % params

        if self.request.headers.get("Content-Type") == "application/json":
            param_dashboard = self.dashboard().parameterized_json(
                self, **params
            )
            param_datasource = []
            for datasource in self.datasources:
                # ugly hack to avoid calling the datasource twice per
                # DashboardPage (once because it's declared in the datasources,
                # used to automatically register the URLSpecs, and once by the
                # javascript facility that will look for configuration changes)
                if datasource.url_name.startswith("datasource_ConfigChanges"):
                    continue

                value = datasource.parameterized_json(self, **params)
                value["data_url"] = self.reverse_url(
                    datasource.url_name, *args
                )
                param_datasource.append(value)

            # tell the frontend how to get the configuration changes, if the
            # DashboardPage provided it
            param_timeline = None
            if self.timeline:
                # Dashboards can specify a subset of arguments to use for the
                # timeline.
                if self.timeline_params:
                    tl_args = [
                        params[prm]
                        for prm in self.params
                        if prm in self.timeline_params
                    ]
                else:
                    tl_args = args
                param_timeline = self.reverse_url(
                    self.timeline.url_name, *tl_args
                )

            last = len(self.breadcrumb) - 1
            breadcrumbs = [
                {
                    "text": item.title,
                    "href": self.reverse_url(
                        item.url_name, *item.url_params.values()
                    ),
                }
                for i, item in enumerate(reversed(self.breadcrumb))
            ] + [
                {
                    "text": item.children_title,
                    "children": (
                        [
                            {
                                "url": self.reverse_url(
                                    child.url_name, *child.url_params.values()
                                ),
                                "title": child.title,
                            }
                            for child in item.children
                        ]
                        if item.children and i == last
                        else None
                    ),
                }
                for i, item in enumerate(reversed(self.breadcrumb))
                if i == last and item.children
            ]

            return self.render_json(
                dict(
                    dashboard=param_dashboard,
                    datasources=param_datasource,
                    timeline=param_timeline,
                    title=title,
                    breadcrumbs=breadcrumbs,
                    handler=self.to_json(),
                )
            )

        return self.render(self.template, title=title)

    @property
    def database(self):
        params = dict(zip(self.params, self.path_args))
        return params.get("database", None)

    @property
    def server(self):
        params = dict(zip(self.params, self.path_args))
        return params.get("server", None)

    @property
    def notify_allowed(self):
        """
        Returns whether the current user is allowed to use the NOTIFY-based
        communication with powa-collector
        """
        conn = self.connect()
        cur = conn.cursor()

        powa_roles = [
            "powa_signal_backend",
            "powa_write_all_data",
            "powa_admin",
        ]
        roles_to_test = []

        # pg 14+ introduced predefined roles
        if conn.server_version >= 140000:
            roles_to_test.append("pg_signal_backend")

        try:
            cur.execute(
                """WITH s(v) AS (
                    SELECT unnest(%s::text[])
                    UNION ALL
                    SELECT rolname
                    FROM {powa}.powa_roles
                    WHERE powa_role = ANY (%s)
                )
                SELECT bool_or(pg_has_role(current_user, v, 'USAGE'))
                    FROM s""",
                (roles_to_test, powa_roles),
            )
        except psycopg2.Error:
            conn.rollback()
            return False

        row = cur.fetchone()

        # should not happen
        if not row:
            return False

        return row[0] is True

    @property
    def breadcrumb(self):
        params = OrderedDict(zip(self.params, self.path_args))
        breadcrumb = self.get_breadcrumb(self, params)
        return breadcrumb


class MetricGroupHandler(AuthHandler):
    """
    Handler for a metric group.
    """

    def initialize(self, datasource, params):
        self.params = params
        self.metric_group = datasource

    def get(self, *params):
        url_params = dict(zip(self.params, params))
        url_query_params = dict(
            (
                (key, value[0].decode("utf8"))
                for key, value in self.request.arguments.items()
            )
        )
        url_params.update(url_query_params)
        url_params = self.add_params(url_params)
        res = self.query
        if res is None:
            query = None
        elif isinstance(res, str):
            query = res
        else:
            query = res[0]
            url_params.update(res[1])
        data = {"data": []}
        if query is not None:
            values = self.execute(query, params=url_params)
            if values is not None:
                data = {
                    "data": [self.process(val, **url_params) for val in values]
                }

        data = self.post_process(data, **url_params)
        self.render_json(data)

    def add_params(self, params):
        return params

    def process(self, val, **kwargs):
        """
        Callback used to process each individual row fetched from the query.

        Arguments:
            handler (tornado.web.RequestHandler):
                the current request handler
            val (psycopg2.RealDictRow):
                the row to process
            kwargs (dict):
                the current url_parameters
        Returns:
            A dictionary containing the processed values.
        """
        return dict(val)

    def post_process(self, data, **kwargs):
        """
        Callback used to process the whole set of rows before returning
        it to the browser.

        Arguments:
            handler (tornado.web.RequestHandler):
                the current request handler
            val (psycopg2.RealDictRow):
                the row to process
            kwargs (dict):
                the current url_parameters
        Returns:
            A dictionary containing the processed values.
        """
        return data


class DataSource(JSONizable):
    """
    Base class for various datasources

    Attributes:
        datasource_handler_cls (type):
            a subclass of RequestHandler used to process this DataSource.

    """

    datasource_handler_cls = None
    data_url = None
    enabled = True

    @classproperty
    def url_name(cls):
        """
        Returns the default url_name for this data source.
        """
        return "datasource_%s" % cls.__name__

    @classproperty
    def parameterized_json(cls, handler, **parms):
        return cls.to_json()


class Metric(JSONizable):
    """
    An indivudal Metric.
    A Metric is an abstraction for a series of data.

    Attributes:
        name: the name of the metric
        label: its label, as displayed in the ui
        type: a type of unit, as referenced in the widget using the type.
              See GraphView.js for example
    """

    def __init__(self, name, label=None, yaxis=None, **kwargs):
        self.name = name
        self.label = label or name.title()
        self.yaxis = yaxis or name
        self._group = None
        self.desc = "No description provided."
        for key, value in kwargs.items():
            setattr(self, key, value)

    def bind(self, group):
        """
        Bind the metric to a metric group. Each metric belong to
        only one MetricGroup.
        """

        if self._group is not None:
            raise ValueError("Already bound to %s" % self._group)
        self._group = group

    def _fqn(self):
        """
        Return the fully qualified name of this metric.
        """
        return "%s.%s" % (self._group.name, self.name)


class Dashboard(JSONizable):
    """
    A Dashboard definition.

    Attributes:
        title (str):
            the dashboard title
        _widgets (list of list):
            A list of rows, with each row containing a list of widgets.
    """

    def __init__(self, title, widgets=None):
        self.title = title
        self._widgets = widgets or []

    def _validate_layout(self):
        """
        Validate that the layout is consistent.
        """
        if not isinstance(self._widgets, list):
            raise ValueError("Widgets must be a list of list of widgets")
        for row in self._widgets:
            if (12 % len(row)) != 0:
                raise ValueError(
                    "Each widget row length must be a "
                    "divisor of 12 (have: %d)" % len(row)
                )

    @property
    def widgets(self):
        """
        Returns this dashboard widgets.
        """
        return self._widgets

    @widgets.setter
    def set_widgets(self, widgets):
        """
        Widgets setter.
        """
        self._validate_layout()
        self._widgets = widgets

    def to_json(self):
        self._validate_layout()
        return {
            "title": self.title,
            "type": "dashboard",
            "widgets": self.widgets,
        }

    def param_widgets(self, _, **params):
        param_rows = []
        for row in self.widgets:
            param_row = []
            for widget in row:
                param_row.append(widget.parameterized_json(_, **params))
            param_rows.append(param_row)
        return param_rows

    def parameterized_json(self, _, **params):
        return {
            "title": self.title % params,
            "type": "dashboard",
            "widgets": self.param_widgets(_, **params),
        }


class Panel(JSONizable):
    def __init__(self, title, widget):
        self.title = title
        self.widget = widget

    def to_json(self):
        return {"title": self.title, "widget": self.widget}

    def parameterized_json(self, _, **args):
        return {
            "title": self.title % args,
            "type": "panel",
            "widget": self.widget.parameterized_json(_, **args),
        }


class TabContainer(JSONizable):
    def __init__(self, title, tabs=None):
        self.title = title
        self.tabs = tabs or []

    def to_json(self):
        return {"title": self.title, "tabs": self.tabs}

    def parameterized_json(self, _, **params):
        tabs = []
        for tab in self.tabs:
            tabs.append(tab.parameterized_json(_, **params))
        return {
            "title": self.title % params,
            "type": "tabcontainer",
            "tabs": tabs,
        }


class Widget(JSONizable):
    """
    Base class for every Widget.
    """

    def parameterized_json(self, _, **params):
        base = params.copy()
        base.update(self.to_json())
        base["title"] = base["title"] % params
        return base


class ContentHandler(AuthHandler):
    """
    Base class for ContentHandlers.

    ContentHandler subclasses are generated on the fly by ContentWidgets,
    when they are registered.
    """

    def initialize(self, datasource=None, params=None):
        self.params = params


class ContentWidget(Widget, DataSource, AuthHandler):
    """
    A widget showing HTML fetched from the server.

    This widget acts as both a Widget and DataSource, since the Data used is
    simplistic.
    """

    datasource_handler_cls = ContentHandler

    def initialize(self, datasource=None, params=None):
        self.params = params

    @classmethod
    def to_json(cls):
        return {
            "data_url": cls.data_url,
            "name": cls.__name__,
            "type": cls.__name__,
            "title": cls.title,
        }

    @classmethod
    def parameterized_json(cls, _, **params):
        return cls.to_json()


class Grid(Widget):
    """
    A rich table Widget, backed by a BackGrid component.

    Attributes:
        columns (list):
            a list of column definitions, excluding metrics
        metrics (list):
            a list of metrics to be included as columns
    """

    def __init__(self, title, columns=None, metrics=None, **kwargs):
        self.title = title
        self.metrics = metrics or []
        self.columns = columns or []
        for key, value in kwargs.items():
            setattr(self, key, value)
        self._validate()

    def _validate(self):
        """
        Validate that the metrics are coherent.
        """
        if len(self.metrics) > 0:
            mg1 = self.metrics[0]._group
            if any(m._group != mg1 for m in self.metrics):
                raise ValueError(
                    "A grid is not allowed to have metrics from different "
                    "groups. (title: %s)" % self.title
                )

    def to_json(self):
        values = self.__dict__.copy()
        values["metrics"] = []
        values["type"] = "grid"
        for metric in self.metrics:
            values["metrics"].append(metric._fqn())
        return values


class Graph(Widget):
    """
    A widget backed by a Rickshaw graph.
    """

    def __init__(
        self,
        title,
        grouper=None,
        url=None,
        axistype="time",
        metrics=None,
        desc=None,
        **kwargs,
    ):
        self.title = title
        self.desc = desc
        self.url = url
        self.grouper = grouper
        self.metrics = metrics or []
        self.axistype = "time"
        # rickshaw default behavior is to stack metrics, but most of our graphs
        # use line renderer with unstacked metrics, so by default unstack
        # metrics, callers will override it if they want
        self.stack = False
        for key, value in kwargs.items():
            setattr(self, key, value)

    def _validate_axis(self, metrics):
        if len(metrics) == 0:
            return
        axis_type = metrics[0].axis_type
        for metric in metrics:
            if metric.axis_type != axis_type:
                raise ValueError(
                    "Some metrics do not have the same x-axis type!"
                )

    def to_json(self):
        values = self.__dict__.copy()
        values["metrics"] = []
        values["type"] = "graph"
        for metric in self.metrics:
            values["metrics"].append(metric._fqn())
        return values


class Declarative(object):
    """
    Base class for declarative classes.
    """

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        global GLOBAL_COUNTER
        self.kwargs.setdefault("_order", GLOBAL_COUNTER)
        GLOBAL_COUNTER += 1


class MetricDef(Declarative):
    """
    A metric definition.
    """

    _cls = Metric


class MetaMetricGroup(type, JSONizable):
    """
    Meta class for Metric Groups.
    This meta class parses its MetricDef attributes, and instantiates
    real Metrics from them.
    """

    def __new__(meta, name, bases, dct):
        dct["metrics"] = {}
        dct["_stubs"] = {}
        if not isinstance(dct.get("name", ""), str):
            raise ValueError("The metric group name must be of type str")
        for base in bases:
            if hasattr(base, "enabled"):
                dct.setdefault("enabled", base.enabled)
            if hasattr(base, "_stubs"):
                for key, stub in base._stubs.items():
                    dct[key] = stub.__class__(*stub.args, **stub.kwargs)
        for key, val in list(dct.items()):
            if isinstance(val, Declarative):
                dct["_stubs"][key] = val
                val.kwargs["name"] = key
                dct[key] = val = val._cls(*val.args, **val.kwargs)
            if isinstance(val, Metric):
                dct.pop(key)
                dct["metrics"][key] = val
        return super(MetaMetricGroup, meta).__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        for metric in dct.get("metrics").values():
            metric.bind(cls)
        super(MetaMetricGroup, cls).__init__(name, bases, dct)

    def __getattr__(cls, key):
        if key not in cls.metrics:
            raise AttributeError
        return cls.metrics.get(key)

    def __hasattr__(cls, key):
        return key in cls.metrics


class MetricGroupDef(with_metaclass(MetaMetricGroup, DataSource)):
    """
    Base class for MetricGroupDef.

    A MetricGroupDef provides syntactic sugar for instantiating MetricGroups.
    """

    _inst = None
    datasource_handler_cls = MetricGroupHandler

    @classmethod
    def to_json(cls):
        values = dict(
            (
                (key, val)
                for key, val in cls.__dict__.items()
                if not key.startswith("_") and not isfunction(val)
            )
        )
        values["type"] = "metric_group"
        values.setdefault("xaxis", "ts")
        values["metrics"] = list(cls.metrics.values())
        values.pop("query", None)
        return values

    @classmethod
    def _get_metrics(cls, handler, **params):
        return cls.metrics

    @classmethod
    def parameterized_json(cls, handler, **params):
        base = cls.to_json()
        base["metrics"] = list(cls._get_metrics(handler, **params).values())
        return base

    @classmethod
    def all(cls, handler=None, **params):
        if handler is None:
            return sorted(cls.metrics.values(), key=attrgetter("_order"))

        return sorted(
            cls._get_metrics(handler, **params).values(),
            key=attrgetter("_order"),
        )

    @classmethod
    def split(cls, handler, splits):
        """
        Returns a list of list containing all the metrics, split according
        to the given list of list of metric names.  The array first element
        contains all the metrics that don't appear in the splits.  For
        instance:
            split_metrics(..., [ ["b", "c"], ["d"] ])
        called on a dict of metrics containing the key "a" to "e" would return
        [[metric_a, metric_e], [metric_b, metric_c], [metric_d]]
        """
        metrics = cls._get_metrics(handler).copy()

        res = []

        i = 0
        res.append({})

        # Distribute the metrics according to the splits
        for split in splits:
            i += 1
            res.append({})
            for k in split:
                res[i][k] = metrics.pop(k)

        # and add what's left in the first position
        res[0] = metrics

        for i in range(0, len(res)):
            res[i] = sorted(res[i].values(), key=attrgetter("_order"))

        return res

    @property
    def metrics(self):
        return self._metrics


class DashboardPage(object):
    """
    A Dashboard page ties together a set of datasources, and a dashboard.

    Attributes:
        template (str):
            the template to use to render the dashboard
        params (list):
            a list of parameter names mapped to groups in the url regexp
        base_url (str):
            the url to this page
        dashboard_handler_cls (RequestHandler):
            the RequestHandler class used to display this page
        datasources (list):
            the list of datasources to include in the page.
    """

    template = "fullpage_dashboard.html"
    params = []
    dashboard_handler_cls = DashboardHandler
    base_url = None
    datasources = []
    parent = None
    timeline = None
    timeline_params = None
    docs_stats_url = (
        "https://powa.readthedocs.io/en/latest/components/stats_extensions/"
    )

    @classmethod
    def url_specs(cls, url_prefix):
        """
        Return the URLSpecs to be register on the application.
        This usually includes one URLSpec for the page itself, and one for
        each datasource.
        """

        url_specs = []
        url_specs.append(
            URLSpec(
                r"%s%s/" % (url_prefix, cls.base_url.strip("/")),
                type(cls.__name__, (cls.dashboard_handler_cls, cls), {}),
                {"template": cls.template, "params": cls.params},
                name=cls.__name__,
            )
        )
        for datasource in cls.datasources:
            if datasource.data_url is None:
                raise KeyError(
                    "A Datasource must have a data_url: %s"
                    % datasource.__name__
                )
            url_specs.append(
                URLSpec(
                    r"%s%s/" % (url_prefix, datasource.data_url.strip("/")),
                    type(
                        datasource.__name__,
                        (datasource, datasource.datasource_handler_cls),
                        dict(datasource.__dict__),
                    ),
                    {"datasource": datasource, "params": cls.params},
                    name=datasource.url_name,
                )
            )
        return url_specs

    @classmethod
    def get_childmenu(cls, handler, params):
        return None

    @classmethod
    def get_selfmenu(cls, handler, params):
        my_params = OrderedDict((key, params.get(key)) for key in cls.params)
        return MenuEntry(cls.title % params, cls.__name__, my_params)

    @classmethod
    def get_breadcrumb(cls, handler, params):
        if getattr(cls, "breadcrum_title", None):
            title = cls.breadcrum_title(handler, params)
        else:
            title = cls.title % params
        entry = MenuEntry(title, cls.__name__, params)
        entry.children_title = (
            cls.title % params if hasattr(cls, "title") else None
        )
        entry.children = cls.get_childmenu(handler, params)
        items = [entry]

        if len(params) > 0:
            parent_params = params.copy()
            parent_params.popitem()
        else:
            parent_params = []

        if (
            cls.parent is not None
            and hasattr(handler, "parent")
            and hasattr(cls.parent, "get_breadcrumb")
        ):
            items.extend(cls.parent.get_breadcrumb(handler, parent_params))
        return items



import os
import sys
from tornado.options import (
    Error,
    define,
    options,
    parse_command_line,
    parse_config_file,
)

SAMPLE_CONFIG_FILE = """
servers={
  'main': {
    'host': 'localhost',
    'port': '5432',
    'database': 'powa'
  }
}
cookie_secret="SUPERSECRET_THAT_YOU_SHOULD_CHANGE"
"""

CONF_LOCATIONS = [
    "/etc/powa-web.conf",
    os.path.expanduser("~/.config/powa-web.conf"),
    os.path.expanduser("~/.powa-web.conf"),
    "./powa-web.conf",
]


define("cookie_secret", type=str, help="Secret key for cookies")
define(
    "cookie_expires_days",
    type=int,
    default=30,
    help="Cookie retention in days",
)
define("port", type=int, default=8888, metavar="port", help="Listen on <port>")
define(
    "address",
    type=str,
    default="0.0.0.0",
    metavar="address",
    help="Listen on <address>",
)
define("config", type=str, help="path to config file")
define("url_prefix", type=str, help="optional prefix URL", default="/")
define(
    "allow_ui_connection",
    type=bool,
    help="Allow UI to connect to databases",
    default=True,
)
define("certfile", type=str, help="Path to certificate file", default=None)
define("keyfile", type=str, help="Path to key file", default=None)


def parse_file(filepath):
    try:
        parse_config_file(filepath)
    except IOError:
        pass
    except Error as e:
        print("Error parsing config file %s:" % filepath)
        print("\t%s" % e)
        sys.exit(-1)


def parse_options():
    define("servers", type=dict, help="Not available from the command line.")
    for possible_config in CONF_LOCATIONS:
        parse_file(possible_config)
    try:
        parse_command_line()
        if options.config:
            parse_file(options.config)
    except Error as e:
        print("Error parsing command line options:")
        print("\t%s" % e)
        sys.exit(1)

    for key in ("servers", "cookie_secret"):
        if getattr(options, key, None) is None:
            print(
                "You should define a server and cookie_secret in your "
                "configuration file."
            )
            print(
                "Place and adapt the following content in one of those "
                "locations:"
                ""
            )
            print("\n\t".join([""] + CONF_LOCATIONS))
            print(SAMPLE_CONFIG_FILE)
            sys.exit(-1)

    if getattr(options, "url_prefix", "") == "":
        setattr(options, "url_prefix", "/")
    elif getattr(options, "url_prefix", "/") != "/":
        prefix = getattr(options, "url_prefix", "/")
        if prefix[0] != "/":
            prefix = "/" + prefix
        if prefix[-1] != "/":
            prefix = prefix + "/"
        setattr(options, "url_prefix", prefix)
    define(
        "index_url",
        type=str,
        default="%sserver/" % getattr(options, "url_prefix", "/"),
    )

    # we used to expect a field named "username", so accept "username" as an
    # alias for "user"
    for key, conf in getattr(options, "servers", {}).items():
        if "username" in conf.keys():
            conf["user"] = conf["username"]
            del conf["username"]



"""
Datasource template used for the various IO pages.
"""

from powa.dashboards import MetricDef, MetricGroupDef
from powa.sql.utils import sum_per_sec
from powa.sql.views_graph import powa_get_io_sample
from powa.sql.views_grid import powa_getiodata


class TemplateIoGraph(MetricGroupDef):
    """
    Template metric group for IO graph.
    """

    xaxis = "ts"
    query_qual = None

    reads = MetricDef(
        label="Reads", type="sizerate", desc="Amount of data read per second"
    )
    read_time = MetricDef(
        label="Read time",
        type="duration",
        desc="Total time spend reading data per second",
    )
    writes = MetricDef(
        label="Write",
        type="sizerate",
        desc="Amount of data written per second",
    )
    write_time = MetricDef(
        label="Write time",
        type="duration",
        desc="Total time spend writing data per second",
    )
    writebacks = MetricDef(
        label="Writebacks",
        type="sizerate",
        desc="Amount of data writeback per second",
    )
    writeback_time = MetricDef(
        label="Writeback time",
        type="duration",
        desc="Total time spend doing writeback per second",
    )
    extends = MetricDef(
        label="Extends",
        type="sizerate",
        desc="Amount of data extended per second",
    )
    extend_time = MetricDef(
        label="Extend time",
        type="duration",
        desc="Total time spend extending relations per second",
    )
    hits = MetricDef(
        label="Hits",
        type="sizerate",
        desc="Amount of data found in shared_buffers per second",
    )
    evictions = MetricDef(
        label="Eviction",
        type="sizerate",
        desc="Amount of data evicted from shared_buffers per second",
    )
    reuses = MetricDef(
        label="Reuses",
        type="sizerate",
        desc="Amount of data reused in shared_buffers per second",
    )
    fsyncs = MetricDef(
        label="Fsyncs", type="sizerate", desc="Blocks flushed per second"
    )
    fsync_time = MetricDef(
        label="Fsync time",
        type="duration",
        desc="Total time spend flushing block per second",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg15 or less
        if pg_version_num is None or pg_version_num < 160000:
            return {}
        return base

    @property
    def query(self):
        query = powa_get_io_sample(self.query_qual)

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("reads"),
            sum_per_sec("read_time"),
            sum_per_sec("writes"),
            sum_per_sec("write_time"),
            sum_per_sec("writebacks"),
            sum_per_sec("writeback_time"),
            sum_per_sec("extends"),
            sum_per_sec("extend_time"),
            sum_per_sec("hits"),
            sum_per_sec("evictions"),
            sum_per_sec("reuses"),
            sum_per_sec("fsyncs"),
            sum_per_sec("fsync_time"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class TemplateIoGrid(MetricGroupDef):
    """
    Template metric group for IO grid.
    """

    axis_type = "category"
    query_qual = None

    reads = MetricDef(
        label="Reads", type="size", desc="Total amount of data read"
    )
    read_time = MetricDef(
        label="Read time",
        type="duration",
        desc="Total amount of time reading data",
    )
    writes = MetricDef(
        label="Writes", type="size", desc="Total amount of data write"
    )
    write_time = MetricDef(
        label="Write time",
        type="duration",
        desc="Total amount of time writing data",
    )
    writebacks = MetricDef(
        label="Writebacks", type="size", desc="Total amount of data writeback"
    )
    writeback_time = MetricDef(
        label="Writeback time",
        type="duration",
        desc="Total amount of time doing data writeback",
    )
    extends = MetricDef(
        label="Extends", type="size", desc="Total amount of data extension"
    )
    extend_time = MetricDef(
        label="Extend time",
        type="duration",
        desc="Total amount of time extending data",
    )
    hits = MetricDef(
        label="Hits", type="size", desc="Total amount of data hit"
    )
    evictions = MetricDef(
        label="Evictions", type="size", desc="Total amount of data evicted"
    )
    reuses = MetricDef(
        label="Reuses", type="size", desc="Total amount of data reused"
    )
    fsyncs = MetricDef(
        label="Fsyncs", type="size", desc="Total amount of data flushed"
    )
    fsync_time = MetricDef(
        label="Fsync time",
        type="duration",
        desc="Total amount of time flushing data",
    )

    @property
    def query(self):
        query = powa_getiodata(self.query_qual)

        return query

    def process(self, val, **kwargs):
        val["backend_type_url"] = self.reverse_url(
            "ByBackendTypeIoOverview", val["srvid"], val["backend_type"]
        )

        val["obj_url"] = self.reverse_url(
            "ByObjIoOverview", val["srvid"], val["object"]
        )

        val["context_url"] = self.reverse_url(
            "ByContextIoOverview", val["srvid"], val["context"]
        )

        return val



"""
Index page presenting an overview of the cluster stats.
"""

from powa.config import ConfigChangesGlobal
from powa.dashboards import (
    Dashboard,
    DashboardPage,
    Graph,
    Grid,
    MetricDef,
    MetricGroupDef,
    TabContainer,
)
from powa.framework import AuthHandler
from powa.io_template import TemplateIoGraph, TemplateIoGrid
from powa.overview import Overview
from powa.sql.utils import (
    block_size,
    byte_per_sec,
    get_ts,
    mulblock,
    sum_per_sec,
    to_epoch,
    total_hit,
    total_read,
    wps,
)
from powa.sql.views_graph import (
    kcache_getstatdata_sample,
    powa_get_all_idx_sample,
    powa_get_all_tbl_sample,
    powa_get_archiver_sample,
    powa_get_bgwriter_sample,
    powa_get_checkpointer_sample,
    powa_get_database_conflicts_sample,
    powa_get_database_sample,
    powa_get_pgsa_sample,
    powa_get_replication_sample,
    powa_get_slru_sample,
    powa_get_subscription_sample,
    powa_get_user_fct_sample,
    powa_get_wal_receiver_sample,
    powa_get_wal_sample,
    powa_getstatdata_sample,
    powa_getwaitdata_sample,
)
from powa.sql.views_grid import (
    powa_getslrudata,
    powa_getstatdata_db,
    powa_getuserfuncdata_db,
    powa_getwaitdata_db,
)
from tornado.web import HTTPError


class ServerSelector(AuthHandler):
    """Page allowing to choose a server."""

    def get(self):
        self.redirect(
            self.reverse_url("ServerOverview", self.get_argument("srvid"))
        )


class ByDatabaseMetricGroup(MetricGroupDef):
    """
    Metric group used by the "by database" grid
    """

    name = "by_database"
    xaxis = "datname"
    data_url = r"/server/(\d+)/metrics/by_databases/"
    axis_type = "category"
    plantime = MetricDef(label="Plantime", type="duration")
    calls = MetricDef(label="#Calls", type="integer", direction="descending")
    runtime = MetricDef(label="Runtime", type="duration")
    avg_runtime = MetricDef(label="Avg runtime", type="duration")
    shared_blks_read = MetricDef(label="Read", type="size")
    shared_blks_hit = MetricDef(label="Hit", type="size")
    shared_blks_dirtied = MetricDef(label="Dirtied", type="size")
    shared_blks_written = MetricDef(label="Written", type="size")
    temp_blks_read = MetricDef(label="Read", type="size")
    temp_blks_written = MetricDef(label="Written", type="size")
    io_time = MetricDef(label="Time", type="duration")
    wal_records = MetricDef(label="#records", type="integer")
    wal_fpi = MetricDef(label="#FPI", type="integer")
    wal_bytes = MetricDef(label="Bytes", type="size")
    jit_functions = MetricDef(label="Functions", type="integer")
    jit_time = MetricDef(label="Time", type="duration")
    params = ["server"]

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.8"
        ):
            for key in ("plantime", "wal_records", "wal_fpi", "wal_bytes"):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.10"
        ):
            base.pop("jit_functions")
            base.pop("jit_time")

        return base

    @property
    def query(self):
        bs = block_size
        inner_query = powa_getstatdata_db("%(server)s")
        from_clause = """({inner_query}) AS sub
        JOIN {{powa}}.powa_databases pd ON pd.oid = sub.dbid
            AND pd.srvid = sub.srvid""".format(inner_query=inner_query)

        cols = [
            "pd.srvid",
            "pd.datname",
            "sum(calls) AS calls",
            "sum(runtime) AS runtime",
            "round(cast(sum(runtime) AS numeric) / greatest(sum(calls), 1), 2) AS avg_runtime",
            mulblock("shared_blks_read", fn="sum"),
            mulblock("shared_blks_hit", fn="sum"),
            mulblock("shared_blks_dirtied", fn="sum"),
            mulblock("shared_blks_written", fn="sum"),
            mulblock("temp_blks_read", fn="sum"),
            mulblock("temp_blks_written", fn="sum"),
            "round(cast(sum(shared_blk_read_time + shared_blk_write_time"
            + " + local_blk_read_time + local_blk_write_time"
            + " + temp_blk_read_time + temp_blk_write_time"
            + ") AS numeric), 2) AS io_time",
        ]

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        ):
            cols.extend(
                [
                    "sum(plantime) AS plantime",
                    "sum(wal_records) AS wal_records",
                    "sum(wal_fpi) AS wal_fpi",
                    "sum(wal_bytes) AS wal_bytes",
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        ):
            cols.extend(
                [
                    "sum(jit_functions) AS jit_functions",
                    "sum(jit_generation_time + jit_inlining_time"
                    + " + jit_optimization_time + jit_emission_time)"
                    + " AS jit_time",
                ]
            )

        return """SELECT {cols}
        FROM {from_clause}
        CROSS JOIN {bs}
        GROUP BY pd.srvid, pd.datname, block_size""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=bs
        )

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url(
            "DatabaseOverview", val["srvid"], val["datname"]
        )
        return val


class ByDatabaseWaitSamplingMetricGroup(MetricGroupDef):
    """
    Metric group used by the "wait sampling by database" grid
    """

    name = "wait_sampling_by_database"
    xaxis = "datname"
    data_url = r"/server/(\d+)/metrics/wait_event_by_databases/"
    axis_type = "category"
    counts = MetricDef(
        label="# of events", type="integer", direction="descending"
    )

    @property
    def query(self):
        inner_query = powa_getwaitdata_db()

        from_clause = """({inner_query}) AS sub
            JOIN {{powa}}.powa_databases pd ON pd.oid = sub.dbid
            AND pd.srvid = sub.srvid""".format(inner_query=inner_query)

        return """SELECT pd.srvid, pd.datname, sub.event_type, sub.event,
            sum(sub.count) AS counts
            FROM {from_clause}
            GROUP BY pd.srvid, pd.datname, sub.event_type, sub.event
            ORDER BY sum(sub.count) DESC""".format(from_clause=from_clause)

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url(
            "DatabaseOverview", val["srvid"], val["datname"]
        )
        return val


class ByDatabaseUserFuncMetricGroup(MetricGroupDef):
    """
    Metric group used by the pg_stat_user_functions grid
    """

    name = "user_functions_by_database"
    xaxis = "datname"
    data_url = r"/server/(\d+)/metrics/user_functions_by_databases/"
    axis_type = "category"
    calls = MetricDef(
        label="# of calls", type="integer", direction="descending"
    )
    total_time = MetricDef(
        label="Cumulated total execution time", type="duration"
    )
    self_time = MetricDef(
        label="Cumulated self execution time", type="duration"
    )

    @property
    def query(self):
        query = powa_getuserfuncdata_db()

        return query + " ORDER BY calls DESC"

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url(
            "DatabaseOverview", val["srvid"], val["datname"]
        )
        return val


class GlobalDatabasesMetricGroup(MetricGroupDef):
    """
    Metric group used by summarized graphs.
    """

    name = "all_databases"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/databases_globals/"
    avg_runtime = MetricDef(
        label="Avg runtime", type="duration", desc="Average query duration"
    )
    calls = MetricDef(
        label="Queries per sec",
        type="number",
        desc="Number of time the query has been executed",
    )
    planload = MetricDef(
        label="Plantime per sec",
        type="duration",
        desc="Total planning duration",
    )
    load = MetricDef(
        label="Runtime per sec",
        type="duration",
        desc="Total duration of queries executed",
    )
    total_blks_hit = MetricDef(
        label="Total hit",
        type="sizerate",
        desc="Amount of data found in shared buffers",
    )
    total_blks_read = MetricDef(
        label="Total read",
        type="sizerate",
        desc="Amount of data found in OS cache or read from disk",
    )
    wal_records = MetricDef(
        label="#Wal records",
        type="integer",
        desc="Number of WAL records generated",
    )
    wal_fpi = MetricDef(
        label="#Wal FPI",
        type="integer",
        desc="Number of WAL full-page images generated",
    )
    wal_bytes = MetricDef(
        label="Wal bytes", type="size", desc="Amount of WAL bytes generated"
    )

    total_sys_hit = MetricDef(
        label="Total system cache hit",
        type="sizerate",
        desc="Amount of data found in OS cache",
    )
    total_disk_read = MetricDef(
        label="Total disk read",
        type="sizerate",
        desc="Amount of data read from disk",
    )
    minflts = MetricDef(
        label="Soft page faults",
        type="number",
        desc="Memory pages not found in the processor's MMU",
    )
    majflts = MetricDef(
        label="Hard page faults",
        type="number",
        desc="Memory pages not found in memory and loaded from storage",
    )
    # not maintained on GNU/Linux, and not available on Windows
    # nswaps = MetricDef(label="Swaps", type="number")
    # msgsnds = MetricDef(label="IPC messages sent", type="number")
    # msgrcvs = MetricDef(label="IPC messages received", type="number")
    # nsignals = MetricDef(label="Signals received", type="number")
    nvcsws = MetricDef(
        label="Voluntary context switches",
        type="number",
        desc="Number of voluntary context switches",
    )
    nivcsws = MetricDef(
        label="Involuntary context switches",
        type="number",
        desc="Number of involuntary context switches",
    )
    jit_functions = MetricDef(
        label="# of JIT functions",
        type="integer",
        desc="Total number of emitted functions",
    )
    jit_generation_time = MetricDef(
        label="JIT generation time",
        type="duration",
        desc="Total time spent generating code",
    )
    jit_inlining_count = MetricDef(
        label="# of JIT inlining",
        type="integer",
        desc="Number of queries where inlining was done",
    )
    jit_inlining_time = MetricDef(
        label="JIT inlining time",
        type="duration",
        desc="Total time spent inlining code",
    )
    jit_optimization_count = MetricDef(
        label="# of JIT optimization",
        type="integer",
        desc="Number of queries where optimization was done",
    )
    jit_optimization_time = MetricDef(
        label="JIT optimization time",
        type="duration",
        desc="Total time spent optimizing code",
    )
    jit_emission_count = MetricDef(
        label="# of JIT emission",
        type="integer",
        desc="Number of queries where emission was done",
    )
    jit_emission_time = MetricDef(
        label="JIT emission time",
        type="duration",
        desc="Total time spent emitting code",
    )
    jit_deform_count = MetricDef(
        label="# of JIT tuple deforming",
        type="integer",
        desc="Number of queries where tuple deforming was done",
    )
    jit_deform_time = MetricDef(
        label="JIT tuple deforming time",
        type="duration",
        desc="Total time spent deforming tuple",
    )
    jit_expr_time = MetricDef(
        label="JIT expression generation time",
        type="duration",
        desc="Total time spent generating expressions",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()
        if not handler.has_extension(params["server"], "pg_stat_kcache"):
            for key in (
                "total_sys_hit",
                "total_disk_read",
                "minflts",
                "majflts",
                # "nswaps", "msgsnds", "msgrcvs", "nsignals",
                "nvcsws",
                "nivcsws",
            ):
                base.pop(key)
        else:
            base.pop("total_blks_read")

        if not handler.has_extension_version(
            params["server"], "pg_stat_statements", "1.8"
        ):
            for key in ("planload", "wal_records", "wal_fpi", "wal_bytes"):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.10"
        ):
            for key in (
                "jit_functions",
                "jit_generation_time",
                "jit_inlining_count",
                "jit_inlining_time",
                "jit_optimization_count",
                "jit_optimization_time",
                "jit_emission_count",
                "jit_emission_time",
            ):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.11"
        ):
            for key in (
                "jit_deform_count",
                "jit_deform_time",
                "jit_expr_time",
            ):
                base.pop(key)

        return base

    @property
    def query(self):
        bs = block_size
        query = powa_getstatdata_sample("db")

        cols = [
            "sub.srvid",
            "extract(epoch FROM ts) AS ts",
            sum_per_sec("calls"),
            "sum(runtime) / greatest(sum(calls), 1) AS avg_runtime",
            sum_per_sec("runtime", alias="load"),
            total_read("sub"),
            total_hit("sub"),
        ]

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        ):
            cols.extend(
                [
                    sum_per_sec("plantime", alias="planload"),
                    sum_per_sec("wal_records"),
                    sum_per_sec("wal_fpi"),
                    sum_per_sec("wal_bytes"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_functions"),
                    sum_per_sec("jit_generation_time"),
                    sum_per_sec("jit_inlining_count"),
                    sum_per_sec("jit_inlining_time"),
                    sum_per_sec("jit_optimization_count"),
                    sum_per_sec("jit_optimization_time"),
                    sum_per_sec("jit_emission_count"),
                    sum_per_sec("jit_emission_time"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_deform_count"),
                    sum_per_sec("jit_deform_time"),
                    sum_per_sec(
                        "jit_generation_time - jit_deform_time",
                        alias="jit_expr_time",
                    ),
                ]
            )

        from_clause = query

        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            from_clause = "({query}) AS sub2".format(query=query)

            # Add system metrics from pg_stat_kcache,
            kcache_query = kcache_getstatdata_sample("db")

            total_sys_hit = (
                "{total_read} - sum(sub.reads)/ {ts} AS total_sys_hit".format(
                    total_read=total_read("sub", True), ts=get_ts()
                )
            )
            total_disk_read = (
                "sum(sub.reads) / " + get_ts() + " AS total_disk_read"
            )
            minflts = sum_per_sec("minflts", prefix="sub")
            majflts = sum_per_sec("majflts", prefix="sub")
            # nswaps = sum_per_sec('nswaps', prefix="sub")
            # msgsnds = sum_per_sec('msgsnds', prefix="sub")
            # msgrcvs = sum_per_sec('msgrcvs', prefix="sub")
            # nsignals = sum_per_sec(.nsignals', prefix="sub")
            nvcsws = sum_per_sec("nvcsws", prefix="sub")
            nivcsws = sum_per_sec("nivcsws", prefix="sub")

            cols.extend(
                [
                    total_sys_hit,
                    total_disk_read,
                    minflts,
                    majflts,
                    # nswaps, msgsnds, msgrcvs, nsignals,
                    nvcsws,
                    nivcsws,
                ]
            )

            from_clause += """
            LEFT JOIN ({kcache_query}) AS kc USING (dbid, ts, srvid)""".format(
                kcache_query=kcache_query
            )

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        CROSS JOIN {bs}
        WHERE sub.calls != 0
        GROUP BY sub.srvid, sub.ts, block_size, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=bs
        )


class GlobalWaitsMetricGroup(MetricGroupDef):
    """Metric group for global wait events graphs."""

    name = "all_databases_waits"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/databases_waits/"
    # pg 9.6 only metrics
    count_lwlocknamed = MetricDef(
        label="Lightweight Named",
        desc="Number of named lightweight lock wait events",
    )
    count_lwlocktranche = MetricDef(
        label="Lightweight Tranche",
        desc="Number of lightweight lock tranche wait events",
    )
    # pg 10+ metrics
    count_lwlock = MetricDef(
        label="Lightweight Lock",
        desc="Number of wait events due to lightweight locks",
    )
    count_lock = MetricDef(
        label="Lock", desc="Number of wait events due to heavyweight locks"
    )
    count_bufferpin = MetricDef(
        label="Buffer pin", desc="Number of wait events due to buffer pin"
    )
    count_activity = MetricDef(
        label="Activity",
        desc="Number of wait events due to postgres"
        " internal processes activity",
    )
    count_client = MetricDef(
        label="Client", desc="Number of wait events due to client activity"
    )
    count_extension = MetricDef(
        label="Extension",
        desc="Number wait events due to third-party extensions",
    )
    count_ipc = MetricDef(
        label="IPC",
        desc="Number of wait events due to inter-process communication",
    )
    count_timeout = MetricDef(
        label="Timeout", desc="Number of wait events due to timeouts"
    )
    count_io = MetricDef(
        label="IO", desc="Number of wait events due to IO operations"
    )

    def prepare(self):
        if not self.has_extension(self.path_args[0], "pg_wait_sampling"):
            raise HTTPError(501, "pg_wait_sampling is not installed")

    @property
    def query(self):
        query = powa_getwaitdata_sample("db")

        cols = [to_epoch("ts", "sub")]

        pg_version_num = self.get_pg_version_num(self.path_args[0])
        # if we can't connect to the remote server, assume pg10 or above
        if pg_version_num is not None and pg_version_num < 100000:
            cols += [
                wps("count_lwlocknamed"),
                wps("count_lwlocktranche"),
                wps("count_lock"),
                wps("count_bufferpin"),
            ]
        else:
            cols += [
                wps("count_lwlock"),
                wps("count_lock"),
                wps("count_bufferpin"),
                wps("count_activity"),
                wps("count_client"),
                wps("count_extension"),
                wps("count_ipc"),
                wps("count_timeout"),
                wps("count_io"),
            ]

        from_clause = "({query}) AS sub".format(query=query)

        return """SELECT {cols}
        FROM {from_clause}
        -- WHERE sub.count IS NOT NULL
        GROUP BY sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause
        )


class GlobalPGSAMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_activity graphs
    """

    name = "pgsa"
    xaxis = "ts"
    desc = "All backends, excluding autovacuum workers, are included"
    data_url = r"/server/(\d+)/metrics/pgsa/"
    backend_xid_age = MetricDef(label="Backend xid age")
    backend_xmin_age = MetricDef(label="Backend xmin age")
    oldest_backend = MetricDef(
        label="Oldest backend",
        type="duration",
        desc="Age of the oldest backend, excluding replication backends",
    )
    oldest_xact = MetricDef(label="Oldest transaction", type="duration")
    oldest_query = MetricDef(label="Oldest query", type="duration")
    nb_idle = MetricDef(label="# of idle connections")
    nb_active = MetricDef(label="# of active connections")
    nb_idle_xact = MetricDef(label="# of idle in transaction connections")
    nb_fastpath = MetricDef(label="# of connections in fastpath function call")
    nb_idle_xact_abort = MetricDef(
        label="# of idle in transaction (aborted) connections"
    )
    nb_disabled = MetricDef(label="# of disabled connections")
    nb_unknown = MetricDef(label="# of connections in unknown state")
    nb_parallel_query = MetricDef(label="# of parallel queries")
    nb_parallel_worker = MetricDef(label="# of parallel workers")

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        remote_pg_ver = handler.get_pg_version_num(handler.path_args[0])
        if remote_pg_ver is not None and remote_pg_ver < 130000:
            for key in ("nb_parallel_query", "nb_parallel_worker"):
                base.pop(key)
        return base

    @property
    def query(self):
        query = powa_get_pgsa_sample()
        # Those greatests are to avoid having negative values returned, see https://github.com/powa-team/powa-archivist/issues/98
        cols = [
            "extract(epoch FROM ts) AS ts",
            "greatest(max(backend_xid_age),0) AS backend_xid_age",
            "greatest(max(backend_xmin_age),0) AS backend_xmin_age",
            "greatest(max(backend_start_age) FILTER (WHERE datid IS NOT NULL),0) AS oldest_backend",
            "greatest(max(xact_start_age) FILTER (WHERE datid IS NOT NULL),0) AS oldest_xact",
            "greatest(max(query_start_age) FILTER (WHERE datid IS NOT NULL AND state !~ 'idle'),0) AS oldest_query",
            "count(*) FILTER (WHERE state = 'idle') AS nb_idle",
            "count(*) FILTER (WHERE state = 'active') AS nb_active",
            "count(*) FILTER (WHERE state = 'idle in transaction') AS nb_idle_xact",
            "count(*) FILTER (WHERE state = 'fastpath function call') AS nb_fastpath",
            "count(*) FILTER (WHERE state = 'idle in transaction (aborted)') AS nb_idle_xact_abort",
            "count(*) FILTER (WHERE state = 'disabled') AS nb_disabled",
            "count(*) FILTER (WHERE state IS NULL) AS nb_unknown",
            "count(DISTINCT leader_pid) AS nb_parallel_query",
            "count(*) FILTER (WHERE leader_pid IS NOT NULL) AS nb_parallel_worker",
        ]

        return """SELECT {cols}
            FROM ({query}) AS sub
            GROUP BY ts
            """.format(cols=", ".join(cols), query=query)


class GlobalArchiverMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_archiver graphs.
    """

    name = "archiver"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/archiver/"
    nb_arch = MetricDef(
        label="# of archived WAL per second",
        type="number",
        desc="Number of WAL archived per second",
    )
    nb_to_arch = MetricDef(
        label="# of WAL to archive",
        type="number",
        desc="Number of WAL that needs to be archived",
    )

    @property
    def query(self):
        query = powa_get_archiver_sample()
        ts = get_ts()

        cols = [
            "extract(epoch FROM ts) AS ts",
            "round((nb_arch::numeric / " + ts + ")::numeric, 2) AS nb_arch",
            "nb_to_arch",
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub""".format(cols=", ".join(cols), from_clause=query)


class GlobalBgwriterMetricGroup(MetricGroupDef):
    """
    Metric group used by bgwriter graphs.
    """

    name = "bgwriter"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/bgwriter/"
    buffers_clean = MetricDef(
        label="Buffers clean",
        type="sizerate",
        desc="Number of buffers written by the background writer",
    )
    maxwritten_clean = MetricDef(
        label="Maxwritten clean",
        type="number",
        desc="Number of times the background writer"
        " stopped a cleaning scan because it had"
        " written too many buffers",
    )
    buffers_backend = MetricDef(
        label="Buffers backend",
        type="sizerate",
        desc="Number of buffers written directly by a backend",
    )
    buffers_backend_fsync = MetricDef(
        label="Buffers backend fsync",
        type="number",
        desc="Number of times a backend had to"
        " execute its own fsync call"
        " (normally the background writer handles"
        " those even when the backend does its"
        " own write",
    )
    buffers_alloc = MetricDef(
        label="Buffers alloc",
        type="sizerate",
        desc="Number of buffers allocated",
    )

    @property
    def query(self):
        bs = block_size
        query = powa_get_bgwriter_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            byte_per_sec("buffers_clean", prefix="sub"),
            sum_per_sec("maxwritten_clean", prefix="sub"),
            byte_per_sec("buffers_backend", prefix="sub"),
            sum_per_sec("buffers_backend_fsync", prefix="sub"),
            byte_per_sec("buffers_alloc", prefix="sub"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        CROSS JOIN {bs}
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, block_size, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=bs
        )


class GlobalCheckpointerMetricGroup(MetricGroupDef):
    """
    Metric group used by bgwriter graphs.
    """

    name = "checkpointer"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/checkpointer/"
    num_timed = MetricDef(
        label="# of scheduled checkpoints",
        type="number",
        desc="Number of scheduled checkpoints that have been performed",
    )
    num_requested = MetricDef(
        label="# of requested checkpoints",
        type="number",
        desc="Number of requested checkpoints that have been performed",
    )
    write_time = MetricDef(
        label="Write time",
        type="duration",
        desc="Total amount of time that has been"
        " spent in the portion of checkpoint"
        " processing where files are written to"
        " disk, in milliseconds",
    )
    sync_time = MetricDef(
        label="Sync time",
        type="duration",
        desc="Total amount of time that has been"
        " spent in the portion of checkpoint"
        " processing where files are synchronized"
        " to disk, in milliseconds",
    )
    buffers_written = MetricDef(
        label="Buffers checkpoint",
        type="sizerate",
        desc="Number of buffers written during checkpoints",
    )

    @property
    def query(self):
        bs = block_size
        query = powa_get_checkpointer_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "sum(sub.num_timed) AS num_timed",
            "sum(sub.num_requested) AS num_requested",
            sum_per_sec("write_time", prefix="sub"),
            sum_per_sec("sync_time", prefix="sub"),
            byte_per_sec("buffers_written", prefix="sub"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        CROSS JOIN {bs}
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, block_size, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=bs
        )


class GlobalIoMetricGroup(TemplateIoGraph):
    """
    Metric group used by pg_stat_io graphs.
    """

    name = "io"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/io/"


class ByAllIoMetricGroup(TemplateIoGrid):
    """
    Metric group used by the pg_stat_io grid, with full detail (by
    backend_type, object and context).
    """

    name = "io_by_all"
    xaxis = "backend_type"
    data_url = r"/server/(\d+)/metrics/io_by_all/"


class GlobalSlruMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_slru graph.
    """

    pass
    name = "slru"
    xaxis = "name"
    data_url = r"/server/(\d+)/metrics/slru/"
    blks_zeroed = MetricDef(
        label="Zeroed",
        type="sizerate",
        desc="Number of blocks zeroed during initializations",
    )
    blks_hit = MetricDef(
        label="Hit",
        type="sizerate",
        desc="Number of times disk blocks were found already"
        " in the SLRU, so that a read was not necessary"
        " (this only includes hits in the SLRU, not the"
        " operating system's file system cache)",
    )
    blks_read = MetricDef(
        label="Read",
        type="sizerate",
        desc="Number of disk blocks read for this SLRU",
    )
    blks_written = MetricDef(
        label="Written",
        type="sizerate",
        desc="Number of disk blocks written for this SLRU",
    )
    blks_exists = MetricDef(
        label="Exists",
        type="sizerate",
        desc="Number of blocks checked for existence for this SLRU",
    )
    flushes = MetricDef(
        label="Flushes",
        type="number",
        desc="Number of flushes of dirty data for this SLRU",
    )
    truncates = MetricDef(
        label="Truncates",
        type="number",
        desc="Number of truncates for this SLRU",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg13 or more
        if pg_version_num is not None and pg_version_num < 130000:
            return {}
        return base

    @property
    def query(self):
        query = powa_get_slru_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("blks_zeroed"),
            sum_per_sec("blks_hit"),
            sum_per_sec("blks_read"),
            sum_per_sec("blks_written"),
            sum_per_sec("blks_exists"),
            sum_per_sec("flushes"),
            sum_per_sec("truncates"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class ByAllSlruMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_slru grid.
    """

    name = "slru_by_all"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/slru_by_all/"
    blks_zeroed = MetricDef(
        label="Zeroed",
        type="size",
        desc="Total number of blocks zeroed during initializations",
    )
    blks_hit = MetricDef(
        label="Hit",
        type="size",
        desc="Total number of times disk blocks were found"
        " already in the SLRU, so that a read was not"
        " necessary (this only includes hits in the SLRU,"
        " not the operating system's file system cache)",
    )
    blks_read = MetricDef(
        label="Read",
        type="size",
        desc="Total number of disk blocks read for this SLRU",
    )
    blks_written = MetricDef(
        label="Written",
        type="size",
        desc="Total number of disk blocks written for this SLRU",
    )
    blks_exists = MetricDef(
        label="Exists",
        type="size",
        desc="Total number of blocks checked for existence for this SLRU",
    )
    flushes = MetricDef(
        label="Flushes",
        type="number",
        desc="Total number of flushes of dirty data for this SLRU",
    )
    truncates = MetricDef(
        label="Truncates",
        type="number",
        desc="Total number of truncates for this SLRU",
    )

    @property
    def query(self):
        query = powa_getslrudata()

        return query

    def process(self, val, **kwargs):
        val["url"] = self.reverse_url(
            "ByNameSlruOverview", val["srvid"], val["name"]
        )

        return val


class GlobalSubMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_subscription(_stats) graphs.
    """

    pass
    name = "subscriptions"
    xaxis = "name"
    data_url = r"/server/(\d+)/metrics/subscriptions/"
    last_msg_lag = MetricDef(
        label="Last message latency",
        type="duration",
        desc="Time spent transmitting the last message"
        " received from origin WAL sender",
    )
    report_lag = MetricDef(
        label="Report lag",
        type="duration",
        desc="Time elapsed since since last reporting of"
        " WAL location to origin WAL sender",
    )
    apply_error_count = MetricDef(
        label="# apply error",
        type="number",
        desc="Total number of times an error occurred while applying changes",
    )
    sync_error_count = MetricDef(
        label="# sync error",
        type="number",
        desc="Total number of times an error"
        " occurred during the inital table"
        " synchronization",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg15 or more
        if pg_version_num is None:
            return base

        # pg_stat_subscription was added in pg10 and pg_stat_subscription_stats
        # was added in pg15.
        # leader_pid was added in pg16 and worker_type in pg17 but we don't use
        # those for now.
        if pg_version_num < 100000:
            return {}
        elif pg_version_num < 150000:
            base.pop("apply_error_count")
            base.pop("sync_error_count")

        return base

    @property
    def query(self):
        query = powa_get_subscription_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "max(last_msg_lag) AS last_msg_lag",
            "max(report_lag) AS report_lag",
            "sum(apply_error_count) AS apply_error_count",
            "sum(sync_error_count) AS sync_error_count",
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class GlobalWalMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_wal graph.
    """

    name = "wal"
    xaxis = "name"
    data_url = r"/server/(\d+)/metrics/wal/"
    wal_records = MetricDef(
        label="# records",
        type="number",
        desc="Total number of WAL records generated",
    )
    wal_fpi = MetricDef(
        label="# fpi",
        type="number",
        desc="Total number of WAL full page images generated",
    )
    wal_bytes = MetricDef(
        label="Generated",
        type="sizerate",
        desc="Total amount of WAL generated in bytes",
    )
    wal_buffers_full = MetricDef(
        label="# buffers full",
        type="number",
        desc="Number of times WAL data was written to"
        " disk because WAL buffers became full",
    )
    wal_write = MetricDef(
        label="# writes",
        type="number",
        desc="Number of times WAL buffers were written out"
        " to disk via XLogWrite request",
    )
    wal_sync = MetricDef(
        label="# sync",
        type="number",
        desc="Number of times WAL files were synced to disk"
        " via issue_xlog_fsync request",
    )
    wal_write_time = MetricDef(
        label="Write time",
        type="duration",
        desc="Total amount of time spent writing WAL buffers"
        " to disk via XLogWrite request",
    )
    wal_sync_time = MetricDef(
        label="Sync time",
        type="duration",
        desc="Total amount of time spent syncing WAL files to"
        " disk via issue_xlog_fsync request",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg14 or more
        if pg_version_num is not None and pg_version_num < 140000:
            return {}
        return base

    @property
    def query(self):
        query = powa_get_wal_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("wal_records"),
            sum_per_sec("wal_fpi"),
            sum_per_sec("wal_bytes"),
            sum_per_sec("wal_buffers_full"),
            sum_per_sec("wal_write"),
            sum_per_sec("wal_sync"),
            sum_per_sec("wal_write_time"),
            sum_per_sec("wal_sync_time"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class GlobalWalReceiverMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_wal_receiver graphs.
    """

    name = "wal_receiver"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/wal_receiver/"
    write_delta = MetricDef(
        label="Write delta",
        type="size",
        desc="Total amount of data received from the"
        " primary but not written yet",
    )
    flush_delta = MetricDef(
        label="Flush delta",
        type="size",
        desc="Total amount of data received and written but not flushed yet",
    )
    last_msg_lag = MetricDef(
        label="Last message latency",
        type="duration",
        desc="Time spent transmitting the last message"
        " received from origin WAL sender",
    )
    report_delta = MetricDef(
        label="Report delta",
        type="size",
        desc="Total amount of data not yet reported to origin WAL sender",
    )
    report_lag = MetricDef(
        label="Report lag",
        type="duration",
        desc="Time elapsed since since last reporting of"
        " WAL location to origin WAL sender",
    )
    received_bytes = MetricDef(
        label="WAL receiver bandwidth",
        type="sizerate",
        desc="Amount of data received from original WAL sender",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg9.6 or more
        if pg_version_num is not None and pg_version_num < 90600:
            return {}
        return base

    @property
    def query(self):
        query = powa_get_wal_receiver_sample()

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "last_received_lsn - written_lsn AS write_delta",
            "written_lsn - flushed_lsn AS flush_delta",
            "extract(epoch FROM ((last_msg_receipt_time - last_msg_send_time) * 1000)) AS last_msg_lag",
            "last_received_lsn - latest_end_lsn AS report_delta",
            "greatest(extract(epoch FROM ((ts - latest_end_time) * 1000)), 0) AS report_lag",
            "received_bytes",
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        --GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class GlobalReplicationMetricGroup(MetricGroupDef):
    """
    Metric group used by pg_stat_replication graphs.
    """

    name = "replication"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/replication/"
    sent_lsn = MetricDef(
        label="Sent delta",
        type="size",
        desc="Maximum amount of data not sent to the replica",
    )
    write_lsn = MetricDef(
        label="Write delta",
        type="size",
        desc="Maximum amount of data sent to the replica but"
        " not written locally",
    )
    flush_lsn = MetricDef(
        label="Flush delta",
        type="size",
        desc="Maximum amount of data written on the replica"
        " but not flushed locally",
    )
    replay_lsn = MetricDef(
        label="Replay delta",
        type="size",
        desc="Maximum amount of data flushed on the replica"
        " but not replayed locally",
    )
    write_lag = MetricDef(
        label="Write lag", type="duration", desc="Maximum write lag in s"
    )
    flush_lag = MetricDef(
        label="Flush lag", type="duration", desc="Maximum flush lag in s"
    )
    replay_lag = MetricDef(
        label="Replay lag", type="duration", desc="Maximum replay lag in s"
    )
    nb_async = MetricDef(
        label="# of async replication connections",
        type="number",
        desc="Number of asynchronous replication connections",
    )
    nb_sync = MetricDef(
        label="# of sync replication connections",
        type="number",
        desc="Number of synchronous replication connections",
    )
    nb_physical_act = MetricDef(
        label="# physical slots (active)",
        type="number",
        desc="Number of active physical replication slots",
    )
    nb_physical_not_act = MetricDef(
        label="# physical slots (disconnected)",
        type="number",
        desc="Number of disconnected physical replication slots",
    )
    nb_logical_act = MetricDef(
        label="# logical slots (active)",
        type="number",
        desc="Number of active logical replication slots",
    )
    nb_logical_not_act = MetricDef(
        label="# logical slots (disconnected)",
        type="number",
        desc="Number of inactive logical replication slots",
    )

    @property
    def query(self):
        query = powa_get_replication_sample()

        from_clause = query

        cols = [
            "extract(epoch FROM sub.ts) AS ts",
            # the datasource retrieves the current lsn first and then the
            # rest of the counters, so it's entirely possible to get a
            # slightly negative number here.  If that happens it just means
            # that there was some activity happening and everything is
            # working as expected.
            "greatest(rep_current_lsn - sent_lsn, 0) AS sent_lsn",
            "sent_lsn - write_lsn AS write_lsn",
            "write_lsn - flush_lsn AS flush_lsn",
            "flush_lsn - replay_lsn AS replay_lsn",
            "coalesce(extract(epoch from write_lag), 0) AS write_lag",
            "coalesce(extract(epoch from flush_lag - write_lag), 0) AS flush_lag",
            "coalesce(extract(epoch from replay_lag - flush_lag), 0) AS replay_lag",
            "nb_async",
            "nb_repl - nb_async AS nb_sync",
            "nb_physical_act",
            "nb_physical_not_act",
            "nb_logical_act",
            "nb_logical_not_act",
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            from_clause=from_clause,
        )


class GlobalDbActivityMetricGroup(MetricGroupDef):
    """
    Metric group used by "Database Activity" graphs.
    """

    name = "all_db_activity"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/all_db_activity/"
    numbackends = MetricDef(
        label="# of connections", desc="Total number of connections"
    )
    xact_commit = MetricDef(
        label="# of commits", desc="Total number of commits per second"
    )
    xact_rollback = MetricDef(
        label="# of rollbacks", desc="Total number of rollbacks per second"
    )
    conflicts = MetricDef(
        label="# of conflicts", desc="Total number of conflicts"
    )
    deadlocks = MetricDef(
        label="# of deadlocks", desc="Total number of deadlocks"
    )
    checksum_failures = MetricDef(
        label="# of checkum_failures", desc="Total number of checkum_failures"
    )
    session_time = MetricDef(
        label="Session time",
        type="duration",
        desc="Total time spent by database sessions per second",
    )
    active_time = MetricDef(
        label="Active time",
        type="duration",
        desc="Total time spent executing SQL statements per second",
    )
    idle_in_transaction_time = MetricDef(
        label="idle in xact time",
        type="duration",
        desc="Total time spent idling while in a transaction per second",
    )
    sessions = MetricDef(
        label="# sessions",
        desc="Total number of sessions established per second",
    )
    sessions_abandoned = MetricDef(
        label="# sessions abandoned",
        desc="Number of database sessions that "
        "were terminated because connection to "
        "the client was lost per second",
    )
    sessions_fatal = MetricDef(
        label="# sessions fatal",
        desc="Number of database sessions that "
        "were terminated by fatal errors",
    )
    sessions_killed = MetricDef(
        label="# sessions killed per second",
        desc="Number of database sessions that "
        "were terminated by operator intervention "
        "per second",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg14 or above
        if pg_version_num is not None and pg_version_num < 140000:
            for key in (
                "session_time",
                "active_time",
                "idle_in_transaction_time",
                "sessions",
                "sessions_abandoned",
                "sessions_fatal",
                "sessions_killed",
            ):
                base.pop(key)
        return base

    @property
    def query(self):
        query = powa_get_database_sample()

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "numbackends",
            wps("xact_commit", do_sum=False),
            wps("xact_rollback", do_sum=False),
            "conflicts",
            "deadlocks",
            "checksum_failures",
            wps("session_time", do_sum=False),
            wps("active_time", do_sum=False),
            wps("idle_in_transaction_time", do_sum=False),
            wps("sessions", do_sum=False),
            wps("sessions_abandoned", do_sum=False),
            wps("sessions_fatal", do_sum=False),
            wps("sessions_killed", do_sum=False),
        ]

        return """SELECT {cols}
        FROM ({query}) sub
        WHERE sub.mesure_interval != '0 s'
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            query=query,
        )


class GlobalDbActivityConflMetricGroup(MetricGroupDef):
    """
    Metric group used by the "Recovery conflicts" graph.
    """

    name = "all_db_activity_conf"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/all_db_activity_conf/"
    confl_tablespace = MetricDef(
        label="Tablespaces",
        desc="Total number of queries that have been"
        " been canceled due to drop tablespaces",
    )
    confl_lock = MetricDef(
        label="Locks",
        desc="Total number of queries that have been "
        "canceled due to lock timeouts",
    )
    confl_snapshot = MetricDef(
        label="Snapshots",
        desc="Total number of queries that have been"
        " canceled due to old snapshots",
    )
    confl_bufferpin = MetricDef(
        label="Pinned buffers",
        desc="Total number of queries that have been"
        " canceled due to pinned buffers",
    )
    confl_deadlock = MetricDef(
        label="Deadlocks",
        desc="Total number of queries that have been"
        " canceled due to deadlocks",
    )
    confl_active_logicalslot = MetricDef(
        label="Logical slots",
        desc="Total number of uses of logical slots that"
        " have canceled due to old snapshots or too"
        " low wal_level on the primary",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()

        pg_version_num = handler.get_pg_version_num(handler.path_args[0])
        # if we can't connect to the remote server, assume pg15 or below
        if pg_version_num is None or pg_version_num < 160000:
            base.pop("confl_active_logicalslot")
        return base

    @property
    def query(self):
        query = powa_get_database_conflicts_sample()

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "confl_tablespace",
            "confl_lock",
            "confl_snapshot",
            "confl_bufferpin",
            "confl_deadlock",
            "confl_active_logicalslot",
        ]

        return """SELECT {cols}
        FROM ({query}) sub
        WHERE sub.mesure_interval != '0 s'
        ORDER BY sub.ts""".format(
            cols=", ".join(cols),
            query=query,
        )


class GlobalAllRelMetricGroup(MetricGroupDef):
    """
    Metric group used by "Database objects" graphs.
    """

    name = "all_relations"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/all_relations/"
    idx_size = MetricDef(
        label="Indexes size", type="size", desc="Size of all indexes"
    )
    tbl_size = MetricDef(
        label="Tables size", type="size", desc="Size of all tables"
    )
    idx_ratio = MetricDef(
        label="Index scans ratio",
        type="percent",
        desc="Ratio of index scan / seq scan",
    )
    idx_scan = MetricDef(
        label="Index scans",
        type="number",
        desc="Number of index scan per second",
    )
    seq_scan = MetricDef(
        label="Sequential scans",
        type="number",
        desc="Number of sequential scan per second",
    )
    n_tup_ins = MetricDef(
        label="Tuples inserted",
        type="number",
        desc="Number of tuples inserted per second",
    )
    n_tup_upd = MetricDef(
        label="Tuples updated",
        type="number",
        desc="Number of tuples updated per second",
    )
    n_tup_hot_upd = MetricDef(
        label="Tuples HOT updated",
        type="number",
        desc="Number of tuples HOT updated per second",
    )
    n_tup_del = MetricDef(
        label="Tuples deleted",
        type="number",
        desc="Number of tuples deleted per second",
    )
    vacuum_count = MetricDef(
        label="# Vacuum", type="number", desc="Number of vacuum per second"
    )
    autovacuum_count = MetricDef(
        label="# Autovacuum",
        type="number",
        desc="Number of autovacuum per second",
    )
    analyze_count = MetricDef(
        label="# Analyze", type="number", desc="Number of analyze per second"
    )
    autoanalyze_count = MetricDef(
        label="# Autoanalyze",
        type="number",
        desc="Number of autoanalyze per second",
    )

    @property
    def query(self):
        query1 = powa_get_all_tbl_sample("db")
        query2 = powa_get_all_idx_sample("db")

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            "sum(tbl_size) AS tbl_size",
            "sum(idx_size) AS idx_size",
            "CASE WHEN sum(sub.idx_scan + sub.seq_scan) = 0"
            " THEN 0"
            " ELSE sum(sub.idx_scan) * 100"
            " / sum(sub.idx_scan + sub.seq_scan)"
            " END AS idx_ratio",
            sum_per_sec("idx_scan", prefix="sub"),
            sum_per_sec("seq_scan", prefix="sub"),
            sum_per_sec("n_tup_ins", prefix="sub"),
            sum_per_sec("n_tup_upd", prefix="sub"),
            sum_per_sec("n_tup_hot_upd", prefix="sub"),
            sum_per_sec("n_tup_del", prefix="sub"),
            sum_per_sec("vacuum_count", prefix="sub"),
            sum_per_sec("autovacuum_count", prefix="sub"),
            sum_per_sec("analyze_count", prefix="sub"),
            sum_per_sec("autoanalyze_count", prefix="sub"),
        ]

        return """SELECT {cols}
        FROM (
            ({query1}) q1
            JOIN (SELECT srvid, dbid, ts, idx_size FROM ({query2})s ) q2
                USING (srvid, dbid, ts)
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), query1=query1, query2=query2
        )


class GlobalUserFctMetricGroup(MetricGroupDef):
    """
    Metric group used by "pg_stat_user_function" graph.
    """

    name = "user_functions"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/user_functions/"
    calls = MetricDef(
        label="# of calls",
        type="number",
        desc="Number of function calls per second",
    )
    total_load = MetricDef(
        label="Total time per sec",
        type="number",
        desc="Total execution time duration",
    )
    self_load = MetricDef(
        label="Self time per sec",
        type="number",
        desc="Self execution time duration",
    )

    @property
    def query(self):
        query = powa_get_user_fct_sample("db")

        from_clause = query

        cols = [
            "sub.srvid",
            "extract(epoch FROM sub.ts) AS ts",
            sum_per_sec("calls"),
            sum_per_sec("total_time", alias="total_load"),
            sum_per_sec("self_time", alias="self_load"),
        ]

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        WHERE sub.mesure_interval != '0 s'
        GROUP BY sub.srvid, sub.ts, sub.mesure_interval
        ORDER BY sub.ts""".format(
            cols=", ".join(cols), from_clause=from_clause
        )


class ServerOverview(DashboardPage):
    """
    ServerOverview dashboard page.
    """

    base_url = r"/server/(\d+)/overview/"
    datasources = [
        GlobalDatabasesMetricGroup,
        ByDatabaseMetricGroup,
        ByDatabaseWaitSamplingMetricGroup,
        GlobalWaitsMetricGroup,
        GlobalBgwriterMetricGroup,
        GlobalCheckpointerMetricGroup,
        GlobalAllRelMetricGroup,
        GlobalUserFctMetricGroup,
        ByDatabaseUserFuncMetricGroup,
        ConfigChangesGlobal,
        GlobalPGSAMetricGroup,
        GlobalArchiverMetricGroup,
        GlobalReplicationMetricGroup,
        GlobalDbActivityMetricGroup,
        GlobalDbActivityConflMetricGroup,
        GlobalIoMetricGroup,
        ByAllIoMetricGroup,
        GlobalSlruMetricGroup,
        ByAllSlruMetricGroup,
        GlobalWalMetricGroup,
        GlobalWalReceiverMetricGroup,
        GlobalSubMetricGroup,
    ]
    params = ["server"]
    parent = Overview
    title = "All databases"
    timeline = ConfigChangesGlobal

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        pg_version_num = self.get_pg_version_num(self.path_args[0])
        pgss18 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        )
        pgss110 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        )
        pgss111 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        )

        all_db_metrics = [
            GlobalDatabasesMetricGroup.avg_runtime,
            GlobalDatabasesMetricGroup.load,
            GlobalDatabasesMetricGroup.calls,
        ]
        if pgss18:
            all_db_metrics.extend([GlobalDatabasesMetricGroup.planload])

        block_graph = Graph(
            "Block access in Bps",
            metrics=[GlobalDatabasesMetricGroup.total_blks_hit],
            color_scheme=None,
        )

        all_db_graphs = [
            [
                Graph(
                    "Query runtime per second (all databases)",
                    metrics=all_db_metrics,
                ),
                block_graph,
            ]
        ]

        if "nb_parallel_query" in GlobalPGSAMetricGroup._get_metrics(self):
            parallel_metrics = ["nb_parallel_query", "nb_parallel_worker"]
        else:
            parallel_metrics = []

        pgsa_metrics = GlobalPGSAMetricGroup.split(
            self,
            [
                [
                    "backend_xid_age",
                    "backend_xmin_age",
                    "oldest_backend",
                    "oldest_xact",
                    "oldest_query",
                ],
                parallel_metrics,
            ],
        )
        all_db_graphs.append(
            [
                Graph(
                    "Global activity (all databases)",
                    metrics=pgsa_metrics[0],
                    desc=GlobalPGSAMetricGroup.desc,
                    renderer="bar",
                    stack=True,
                )
            ]
        )
        if len(pgsa_metrics[2]) > 0:
            all_db_graphs[1].append(
                Graph(
                    "Parallel query (all databases)", metrics=pgsa_metrics[2]
                )
            )
        all_db_graphs[1].append(
            Graph(
                "Backend age (all databases)",
                metrics=pgsa_metrics[1],
                desc=GlobalPGSAMetricGroup.desc,
            )
        )

        graphs_dash = [Dashboard("General Overview", all_db_graphs)]
        graphs = [TabContainer("All databases", graphs_dash)]

        # Add WALs graphs

        # if we can't connect to the remote server, assume pg14 or above
        if pg_version_num is None or pg_version_num >= 140000:
            wal_metrics = GlobalWalMetricGroup.split(
                self, [["wal_write_time", "wal_sync_time"]]
            )
            wals_graphs = [
                [
                    Graph(
                        "WAL activity",
                        url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-WAL-VIEW",
                        metrics=wal_metrics[0],
                    ),
                    Graph(
                        "WAL timing",
                        url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-WAL-VIEW",
                        metrics=wal_metrics[1],
                    ),
                ]
            ]
            graphs_dash.append(Dashboard("WALs", wals_graphs))
        elif pgss18:
            wals_graphs = [
                [
                    Graph(
                        "WAL activity",
                        metrics=[
                            GlobalDatabasesMetricGroup.wal_records,
                            GlobalDatabasesMetricGroup.wal_fpi,
                            GlobalDatabasesMetricGroup.wal_bytes,
                        ],
                    ),
                ]
            ]
            graphs_dash.append(Dashboard("WALs", wals_graphs))

        # Add JIT graphs
        if pgss110:
            jit_tim = [
                GlobalDatabasesMetricGroup.jit_inlining_time,
                GlobalDatabasesMetricGroup.jit_optimization_time,
                GlobalDatabasesMetricGroup.jit_emission_time,
            ]

            if pgss111:
                jit_tim.extend(
                    [
                        GlobalDatabasesMetricGroup.jit_deform_time,
                        GlobalDatabasesMetricGroup.jit_expr_time,
                    ]
                )
            else:
                jit_tim.append(GlobalDatabasesMetricGroup.jit_generation_time)

            jit_cnt = [
                GlobalDatabasesMetricGroup.jit_functions,
                GlobalDatabasesMetricGroup.jit_inlining_count,
                GlobalDatabasesMetricGroup.jit_optimization_count,
                GlobalDatabasesMetricGroup.jit_emission_count,
            ]

            if pgss111:
                jit_cnt.append(GlobalDatabasesMetricGroup.jit_deform_count)

            jit_graphs = [
                [Graph("JIT timing", metrics=jit_tim, stack=True)],
                [Graph("JIT scheduling", metrics=jit_cnt)],
            ]

            graphs_dash.append(Dashboard("JIT", jit_graphs))

        # Add pg_stat_bgwriter / pg_stat_checkpointer graphs
        bgw_graphs = [
            [
                Graph(
                    "Checkpointer scheduling",
                    metrics=[
                        GlobalCheckpointerMetricGroup.num_timed,
                        GlobalCheckpointerMetricGroup.num_requested,
                    ],
                ),
                Graph(
                    "Checkpointer activity",
                    metrics=[
                        GlobalCheckpointerMetricGroup.write_time,
                        GlobalCheckpointerMetricGroup.sync_time,
                        GlobalCheckpointerMetricGroup.buffers_written,
                    ],
                ),
            ],
            [
                Graph(
                    "Background writer",
                    metrics=[
                        GlobalBgwriterMetricGroup.buffers_clean,
                        GlobalBgwriterMetricGroup.maxwritten_clean,
                        GlobalBgwriterMetricGroup.buffers_alloc,
                    ],
                ),
                Graph(
                    "Backends",
                    metrics=[
                        GlobalBgwriterMetricGroup.buffers_backend,
                        GlobalBgwriterMetricGroup.buffers_backend_fsync,
                    ],
                ),
            ],
        ]
        graphs_dash.append(
            Dashboard("Background Writer / Checkpointer", bgw_graphs)
        )

        # Add archiver / replication graphs
        sub_metrics = GlobalSubMetricGroup.split(
            self, [["last_msg_lag", "report_lag"]]
        )
        sub_graphs = []

        sub_graphs.append(
            Graph(
                "Subscriptions message & report",
                url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-SUBSCRIPTION",
                metrics=sub_metrics[1],
            )
        )

        if len(sub_metrics[0]) > 0:
            sub_graphs.append(
                Graph(
                    "Subscriptions errors",
                    url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-SUBSCRIPTION-STATS",
                    metrics=sub_metrics[0],
                )
            )

        arch_graphs = [
            [
                Graph("Archiver", metrics=GlobalArchiverMetricGroup.all(self)),
                Graph(
                    "Replication connections",
                    metrics=[
                        GlobalReplicationMetricGroup.nb_async,
                        GlobalReplicationMetricGroup.nb_sync,
                    ],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "Replication slots",
                    metrics=[
                        GlobalReplicationMetricGroup.nb_physical_act,
                        GlobalReplicationMetricGroup.nb_physical_not_act,
                        GlobalReplicationMetricGroup.nb_logical_act,
                        GlobalReplicationMetricGroup.nb_logical_not_act,
                    ],
                    renderer="bar",
                    stack=True,
                ),
            ],
            [
                Graph(
                    "Replication delta in B",
                    metrics=[
                        GlobalReplicationMetricGroup.sent_lsn,
                        GlobalReplicationMetricGroup.write_lsn,
                        GlobalReplicationMetricGroup.flush_lsn,
                        GlobalReplicationMetricGroup.replay_lsn,
                    ],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "Replication lag in s",
                    metrics=[
                        GlobalReplicationMetricGroup.write_lag,
                        GlobalReplicationMetricGroup.flush_lag,
                        GlobalReplicationMetricGroup.replay_lag,
                    ],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "Recovery conflicts",
                    stack=True,
                    metrics=GlobalDbActivityConflMetricGroup.all(self),
                    # metrics=[GlobalDbActivityConflMetricGroup.confl_tablespace]
                ),
            ],
            [
                Graph(
                    "WAL receiver bandwidth",
                    metrics=[GlobalWalReceiverMetricGroup.received_bytes],
                ),
                Graph(
                    "WAL receiver Delta",
                    url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-WAL-RECEIVER-VIEW",
                    metrics=[
                        GlobalWalReceiverMetricGroup.write_delta,
                        GlobalWalReceiverMetricGroup.flush_delta,
                    ],
                    renderer="bar",
                    stack=True,
                ),
            ],
            [
                Graph(
                    "WAL receiver message",
                    url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-WAL-RECEIVER-VIEW",
                    metrics=[GlobalWalReceiverMetricGroup.last_msg_lag],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "WAL receiver report",
                    url="https://www.postgresql.org/docs/current/monitoring-stats.html#MONITORING-PG-STAT-WAL-RECEIVER-VIEW",
                    metrics=[
                        GlobalWalReceiverMetricGroup.report_delta,
                        GlobalWalReceiverMetricGroup.report_lag,
                    ],
                    renderer="bar",
                    stack=True,
                ),
            ],
            sub_graphs,
        ]
        graphs_dash.append(Dashboard("Archiver / Replication", arch_graphs))

        # Add pg_stat_database graphs
        global_db_graphs = [
            [
                Graph(
                    "Transactions per second",
                    metrics=[
                        GlobalDbActivityMetricGroup.xact_commit,
                        GlobalDbActivityMetricGroup.xact_rollback,
                    ],
                    renderer="bar",
                    stack=True,
                ),
                Graph(
                    "Conflicts & deadlocks",
                    metrics=[
                        GlobalDbActivityMetricGroup.conflicts,
                        GlobalDbActivityMetricGroup.deadlocks,
                    ],
                ),
            ]
        ]
        if "sessions" in GlobalDbActivityMetricGroup._get_metrics(self):
            global_db_graphs.append(
                [
                    Graph(
                        "Cumulated time per second",
                        metrics=[
                            GlobalDbActivityMetricGroup.session_time,
                            GlobalDbActivityMetricGroup.active_time,
                            GlobalDbActivityMetricGroup.idle_in_transaction_time,
                        ],
                    ),
                    Graph(
                        "Sessions per second",
                        metrics=[
                            GlobalDbActivityMetricGroup.sessions,
                            GlobalDbActivityMetricGroup.sessions_abandoned,
                            GlobalDbActivityMetricGroup.sessions_fatal,
                            GlobalDbActivityMetricGroup.sessions_killed,
                        ],
                    ),
                ]
            )
        graphs_dash.append(Dashboard("Database activity", global_db_graphs))

        # Add powa_stat_all_relations graphs
        all_rel_graphs = [
            [
                Graph(
                    "Access pattern",
                    metrics=[
                        GlobalAllRelMetricGroup.seq_scan,
                        GlobalAllRelMetricGroup.idx_scan,
                        GlobalAllRelMetricGroup.idx_ratio,
                    ],
                ),
                Graph(
                    "DML activity",
                    metrics=[
                        GlobalAllRelMetricGroup.n_tup_del,
                        GlobalAllRelMetricGroup.n_tup_hot_upd,
                        GlobalAllRelMetricGroup.n_tup_upd,
                        GlobalAllRelMetricGroup.n_tup_ins,
                    ],
                ),
            ],
            [
                Graph(
                    "Vacuum activity",
                    metrics=[
                        GlobalAllRelMetricGroup.autoanalyze_count,
                        GlobalAllRelMetricGroup.analyze_count,
                        GlobalAllRelMetricGroup.autovacuum_count,
                        GlobalAllRelMetricGroup.vacuum_count,
                    ],
                ),
                Graph(
                    "Object size",
                    metrics=[
                        GlobalAllRelMetricGroup.tbl_size,
                        GlobalAllRelMetricGroup.idx_size,
                    ],
                    renderer="bar",
                    stack=True,
                    color_scheme=["#73c03a", "#65b9ac"],
                ),
            ],
        ]
        graphs_dash.append(Dashboard("Database Objects", all_rel_graphs))

        # Add powa_stat_user_functions graphs
        user_fct_graph = [
            Graph(
                "User functions activity",
                metrics=GlobalUserFctMetricGroup.all(self),
            )
        ]
        user_fct_grid = [
            Grid(
                "User functions activity for all databases",
                columns=[
                    {"name": "datname", "label": "Database", "url_attr": "url"}
                ],
                metrics=ByDatabaseUserFuncMetricGroup.all(self),
            )
        ]
        graphs_dash.append(
            Dashboard("User functions", [user_fct_graph, user_fct_grid])
        )

        sys_graphs = []
        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            block_graph.metrics.insert(
                0, GlobalDatabasesMetricGroup.total_sys_hit
            )
            block_graph.metrics.insert(
                0, GlobalDatabasesMetricGroup.total_disk_read
            )
            block_graph.color_scheme = ["#cb513a", "#65b9ac", "#73c03a"]

            sys_graphs.append(
                [
                    Graph(
                        "System resources (events per sec)",
                        url=self.docs_stats_url + "pg_stat_kcache.html",
                        metrics=[
                            GlobalDatabasesMetricGroup.majflts,
                            GlobalDatabasesMetricGroup.minflts,
                            # GlobalDatabasesMetricGroup.nswaps,
                            # GlobalDatabasesMetricGroup.msgsnds,
                            # GlobalDatabasesMetricGroup.msgrcvs,
                            # GlobalDatabasesMetricGroup.nsignals,
                            GlobalDatabasesMetricGroup.nvcsws,
                            GlobalDatabasesMetricGroup.nivcsws,
                        ],
                    )
                ]
            )

        else:
            block_graph.metrics.insert(
                0, GlobalDatabasesMetricGroup.total_blks_read
            )
            block_graph.color_scheme = ["#cb513a", "#73c03a"]

        # Add pg_stat_io graphs
        if "reads" in GlobalIoMetricGroup._get_metrics(self):
            io_metrics = GlobalIoMetricGroup.split(
                self,
                [
                    ["reads", "writes", "writebacks", "extends", "fsyncs"],
                    ["hits", "evictions", "reuses"],
                ],
            )
            sys_graphs.append(
                [
                    Graph(
                        "IO blocks",
                        metrics=io_metrics[1],
                    ),
                    Graph(
                        "IO timing",
                        metrics=io_metrics[0],
                    ),
                    Graph(
                        "IO misc",
                        metrics=io_metrics[2],
                    ),
                ]
            )

            sys_graphs.append(
                [
                    Grid(
                        "IO summary",
                        columns=[
                            {
                                "name": "backend_type",
                                "label": "Backend Type",
                                "url_attr": "backend_type_url",
                            },
                            {
                                "name": "object",
                                "label": "Object Type",
                                "url_attr": "obj_url",
                            },
                            {
                                "name": "context",
                                "label": "Context",
                                "url_attr": "context_url",
                            },
                        ],
                        metrics=ByAllIoMetricGroup.all(),
                    )
                ]
            )

        if len(sys_graphs) != 0:
            graphs_dash.append(Dashboard("System resources", sys_graphs))

        # if we can't connect to the remote server, assume pg13 or above
        if pg_version_num is None or pg_version_num >= 130000:
            graphs_dash.append(
                Dashboard(
                    "SLRU",
                    [
                        [
                            Graph(
                                "All SLRUs (per second)",
                                metrics=GlobalSlruMetricGroup.all(self),
                            )
                        ],
                        [
                            Grid(
                                "All SLRUs",
                                columns=[
                                    {
                                        "name": "name",
                                        "label": "SLRU name",
                                        "url_attr": "url",
                                    }
                                ],
                                metrics=ByAllSlruMetricGroup.all(self),
                            )
                        ],
                    ],
                )
            )

        if self.has_extension(self.path_args[0], "pg_wait_sampling"):
            metrics = None
            # if we can't connect to the remote server, assume pg10 or above
            if pg_version_num is not None and pg_version_num < 100000:
                metrics = [
                    GlobalWaitsMetricGroup.count_lwlocknamed,
                    GlobalWaitsMetricGroup.count_lwlocktranche,
                    GlobalWaitsMetricGroup.count_lock,
                    GlobalWaitsMetricGroup.count_bufferpin,
                ]
            else:
                metrics = [
                    GlobalWaitsMetricGroup.count_lwlock,
                    GlobalWaitsMetricGroup.count_lock,
                    GlobalWaitsMetricGroup.count_bufferpin,
                    GlobalWaitsMetricGroup.count_activity,
                    GlobalWaitsMetricGroup.count_client,
                    GlobalWaitsMetricGroup.count_extension,
                    GlobalWaitsMetricGroup.count_ipc,
                    GlobalWaitsMetricGroup.count_timeout,
                    GlobalWaitsMetricGroup.count_io,
                ]

            graphs_dash.append(
                Dashboard(
                    "Wait Events",
                    [
                        [
                            Graph(
                                "Wait Events (per second)",
                                url=self.docs_stats_url
                                + "pg_wait_sampling.html",
                                metrics=metrics,
                            )
                        ]
                    ],
                )
            )

        toprow = [
            {
                # database
            }
        ]

        if pgss18:
            toprow.extend(
                [
                    {
                        # plan time
                    }
                ]
            )

        toprow.extend(
            [
                {"name": "Execution", "colspan": 3},
                {
                    "name": "Blocks",
                    "colspan": 4,
                },
                {"name": "Temp blocks", "colspan": 2},
                {
                    "name": "I/O",
                },
            ]
        )

        if pgss18:
            toprow.extend([{"name": "WAL", "colspan": 3}])

        if pgss110:
            toprow.extend([{"name": "JIT", "colspan": 2}])

        dashes = [
            graphs,
            [
                Grid(
                    "Details for all databases",
                    toprow=toprow,
                    columns=[
                        {
                            "name": "datname",
                            "label": "Database",
                            "url_attr": "url",
                        }
                    ],
                    metrics=ByDatabaseMetricGroup.all(self),
                )
            ],
        ]

        if self.has_extension(self.path_args[0], "pg_wait_sampling"):
            dashes.append(
                [
                    Grid(
                        "Wait events for all databases",
                        url=self.docs_stats_url + "pg_wait_sampling.html",
                        columns=[
                            {
                                "name": "datname",
                                "label": "Database",
                                "url_attr": "url",
                            },
                            {
                                "name": "event_type",
                                "label": "Event Type",
                            },
                            {
                                "name": "event",
                                "label": "Event",
                            },
                        ],
                        metrics=ByDatabaseWaitSamplingMetricGroup.all(),
                    )
                ]
            )

        self._dashboard = Dashboard("All databases", dashes)
        return self._dashboard

    @classmethod
    def breadcrum_title(cls, handler, param):
        return handler.deparse_srvid(param["server"])

    @classmethod
    def get_childmenu(cls, handler, params):
        from powa.database import DatabaseOverview

        children = []
        for d in list(handler.get_databases(params["server"])):
            new_params = params.copy()
            new_params["database"] = d
            children.append(DatabaseOverview.get_selfmenu(handler, new_params))
        return children



import sys
from setuptools import find_packages, setup

__VERSION__ = None

with open("powa/__init__.py") as f:
    for line in f:
        if line.startswith("__VERSION__"):
            __VERSION__ = line.split("=")[1].replace('"', "").strip()


requires = ["tornado>=2.0", "psycopg2"]

# include ordereddict for python2.6
if sys.version_info < (2, 7, 0):
    requires.append("ordereddict")


setup(
    name="powa-web",
    version=__VERSION__,
    author="powa-team",
    license="Postgresql",
    packages=find_packages(),
    install_requires=requires,
    include_package_data=True,
    url="https://powa.readthedocs.io/",
    description="A User Interface for the PoWA project",
    long_description="See https://powa.readthedocs.io/",
    scripts=["powa-web"],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: System Administrators",
        "Intended Audience :: End Users/Desktop",
        "License :: Other/Proprietary License",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Topic :: Database :: Front-Ends",
    ],
)



"""
Dashboard for the by-query page.
"""

from powa.config import ConfigChangesQuery
from powa.dashboards import (
    ContentWidget,
    Dashboard,
    DashboardPage,
    Graph,
    Grid,
    MetricDef,
    MetricGroupDef,
    TabContainer,
)
from powa.database import DatabaseOverview
from powa.sql import (
    get_any_sample_query,
    get_hypoplans,
    get_plans,
    possible_indexes,
    qualstat_get_figures,
    resolve_quals,
)
from powa.sql.utils import block_size, byte_per_sec, sum_per_sec, to_epoch, wps
from powa.sql.views import QUALSTAT_FILTER_RATIO, qualstat_getstatdata
from powa.sql.views_graph import (
    kcache_getstatdata_sample,
    powa_getstatdata_sample,
    powa_getwaitdata_sample,
)
from powa.sql.views_grid import (
    powa_getstatdata_detailed_db,
    powa_getwaitdata_detailed_db,
)
from psycopg2 import Error
from tornado.web import HTTPError


class QueryOverviewMetricGroup(MetricGroupDef):
    """
    Metric Group for the graphs on the by query page.
    """

    name = "query_overview"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)"
    rows = MetricDef(
        label="#Rows",
        desc="Sum of the number of rows returned by the query per second",
    )
    calls = MetricDef(
        label="#Calls",
        desc="Number of time the query has been executed per second",
    )
    shared_blks_read = MetricDef(
        label="Shared read",
        type="sizerate",
        desc="Amount of data found in OS cache or read from disk",
    )
    shared_blks_hit = MetricDef(
        label="Shared hit",
        type="sizerate",
        desc="Amount of data found in shared buffers",
    )
    shared_blks_dirtied = MetricDef(
        label="Shared dirtied",
        type="sizerate",
        desc="Amount of data modified in shared buffers",
    )
    shared_blks_written = MetricDef(
        label="Shared written",
        type="sizerate",
        desc="Amount of shared buffers written to disk",
    )
    local_blks_read = MetricDef(
        label="Local read",
        type="sizerate",
        desc="Amount of local buffers found from OS cache or read from disk",
    )
    local_blks_hit = MetricDef(
        label="Local hit",
        type="sizerate",
        desc="Amount of local buffers found in shared buffers",
    )
    local_blks_dirtied = MetricDef(
        label="Local dirtied",
        type="sizerate",
        desc="Amount of data modified in local buffers",
    )
    local_blks_written = MetricDef(
        label="Local written",
        type="sizerate",
        desc="Amount of local buffers written to disk",
    )
    temp_blks_read = MetricDef(
        label="Temp read",
        type="sizerate",
        desc="Amount of data read from temporary file",
    )
    temp_blks_written = MetricDef(
        label="Temp written",
        type="sizerate",
        desc="Amount of data written to temporary file",
    )
    shared_blk_read_time = MetricDef(
        label="Shared Read time",
        type="duration",
        desc="Time spent reading shared blocks",
    )
    shared_blk_write_time = MetricDef(
        label="Shared Write time",
        type="duration",
        desc="Time spent shared blocks",
    )
    local_blk_read_time = MetricDef(
        label="Local Read time",
        type="duration",
        desc="Time spent reading temp table blocks",
    )
    local_blk_write_time = MetricDef(
        label="Local Write time",
        type="duration",
        desc="Time spent temp table blocks",
    )
    temp_blk_read_time = MetricDef(
        label="Temp Read time",
        type="duration",
        desc="Time spent reading temp file blocks",
    )
    temp_blk_write_time = MetricDef(
        label="Temp Write time",
        type="duration",
        desc="Time spent temp file blocks",
    )
    avg_plantime = MetricDef(
        label="Avg plantime",
        type="duration",
        desc="Average query planning duration",
    )
    avg_runtime = MetricDef(
        label="Avg runtime", type="duration", desc="Average query duration"
    )
    hit_ratio = MetricDef(
        label="Shared buffers hit ratio",
        type="percent",
        desc="Percentage of data found in shared buffers",
    )
    miss_ratio = MetricDef(
        label="Shared buffers miss ratio",
        type="percent",
        desc="Percentage of data found in OS cache or read from disk",
    )
    wal_records = MetricDef(
        label="#Wal records",
        type="integer",
        desc="Amount of WAL records generated",
    )
    wal_fpi = MetricDef(
        label="#Wal FPI",
        type="integer",
        desc="Amount of WAL full-page images generated",
    )
    wal_bytes = MetricDef(
        label="Wal bytes", type="size", desc="Amount of WAL bytes generated"
    )

    reads = MetricDef(
        label="Physical read",
        type="sizerate",
        desc="Amount of data read from disk",
    )
    writes = MetricDef(
        label="Physical writes",
        type="sizerate",
        desc="Amount of data written to disk",
    )
    user_time = MetricDef(
        label="CPU user time / Query time",
        type="percent",
        desc="CPU time spent executing the query",
    )
    system_time = MetricDef(
        label="CPU system time / Query time",
        type="percent",
        desc="CPU time used by the OS",
    )
    other_time = MetricDef(
        label="CPU other time / Query time",
        type="percent",
        desc="Time spent otherwise",
    )
    sys_hit_ratio = MetricDef(
        label="System cache hit ratio",
        type="percent",
        desc="Percentage of data found in OS cache",
    )
    disk_hit_ratio = MetricDef(
        label="Disk hit ratio",
        type="percent",
        desc="Percentage of data read from disk",
    )
    minflts = MetricDef(
        label="Soft page faults",
        type="number",
        desc="Memory pages not found in the processor's MMU",
    )
    majflts = MetricDef(
        label="Hard page faults",
        type="number",
        desc="Memory pages not found in memory and loaded from storage",
    )
    # not maintained on GNU/Linux, and not available on Windows
    # nswaps = MetricDef(label="Swaps", type="number")
    # msgsnds = MetricDef(label="IPC messages sent", type="number")
    # msgrcvs = MetricDef(label="IPC messages received", type="number")
    # nsignals = MetricDef(label="Signals received", type="number")
    nvcsws = MetricDef(
        label="Voluntary context switches",
        type="number",
        desc="Number of voluntary context switches",
    )
    nivcsws = MetricDef(
        label="Involuntary context switches",
        type="number",
        desc="Number of involuntary context switches",
    )
    jit_functions = MetricDef(
        label="# of JIT functions",
        type="integer",
        desc="Total number of emitted functions",
    )
    jit_generation_time = MetricDef(
        label="JIT generation time",
        type="duration",
        desc="Total time spent generating code",
    )
    jit_inlining_count = MetricDef(
        label="# of JIT inlining",
        type="integer",
        desc="Number of queries where inlining was done",
    )
    jit_inlining_time = MetricDef(
        label="JIT inlining time",
        type="duration",
        desc="Total time spent inlining code",
    )
    jit_optimization_count = MetricDef(
        label="# of JIT optimization",
        type="integer",
        desc="Number of queries where optimization was done",
    )
    jit_optimization_time = MetricDef(
        label="JIT optimization time",
        type="duration",
        desc="Total time spent optimizing code",
    )
    jit_emission_count = MetricDef(
        label="# of JIT emission",
        type="integer",
        desc="Number of queries where emission was done",
    )
    jit_emission_time = MetricDef(
        label="JIT emission time",
        type="duration",
        desc="Total time spent emitting code",
    )
    jit_deform_count = MetricDef(
        label="# of JIT tuple deforming",
        type="integer",
        desc="Number of queries where tuple deforming was done",
    )
    jit_deform_time = MetricDef(
        label="JIT tuple deforming time",
        type="duration",
        desc="Total time spent deforming tuple",
    )
    jit_expr_time = MetricDef(
        label="JIT expression generation time",
        type="duration",
        desc="Total time spent generating expressions",
    )

    @classmethod
    def _get_metrics(cls, handler, **params):
        base = cls.metrics.copy()
        if not handler.has_extension(params["server"], "pg_stat_kcache"):
            for key in (
                "reads",
                "writes",
                "user_time",
                "system_time",
                "other_time",
                "sys_hit_ratio",
                "disk_hit_ratio",
                "minflts",
                "majflts",
                # "nswaps", "msgsnds", "msgrcvs", "nsignals",
                "nvcsws",
                "nivcsws",
            ):
                base.pop(key)
        else:
            base.pop("miss_ratio")

        if not handler.has_extension_version(
            params["server"], "pg_stat_statements", "1.8"
        ):
            for key in ("avg_plantime", "wal_records", "wal_fpi", "wal_bytes"):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.10"
        ):
            for key in (
                "jit_functions",
                "jit_generation_time",
                "jit_inlining_count",
                "jit_inlining_time",
                "jit_optimization_count",
                "jit_optimization_time",
                "jit_emission_count",
                "jit_emission_time",
            ):
                base.pop(key)

        if not handler.has_extension_version(
            handler.path_args[0], "pg_stat_statements", "1.11"
        ):
            for key in (
                "jit_deform_count",
                "jit_deform_time",
                "jit_expr_time",
            ):
                base.pop(key)

        return base

    @property
    def query(self):
        query = powa_getstatdata_sample(
            "query", ["datname = %(database)s", "queryid = %(query)s"]
        )

        total_blocks = "(sum(shared_blks_read) + sum(shared_blks_hit))"

        cols = [
            to_epoch("ts"),
            sum_per_sec("rows"),
            sum_per_sec("calls"),
            "CASE WHEN {total_blocks} = 0 THEN 0 ELSE"
            " sum(shared_blks_hit)::numeric * 100 / {total_blocks}"
            "END AS hit_ratio".format(total_blocks=total_blocks),
            byte_per_sec("shared_blks_read"),
            byte_per_sec("shared_blks_hit"),
            byte_per_sec("shared_blks_dirtied"),
            byte_per_sec("shared_blks_written"),
            byte_per_sec("local_blks_read"),
            byte_per_sec("local_blks_hit"),
            byte_per_sec("local_blks_dirtied"),
            byte_per_sec("local_blks_written"),
            byte_per_sec("temp_blks_read"),
            byte_per_sec("temp_blks_written"),
            sum_per_sec("shared_blk_read_time"),
            sum_per_sec("shared_blk_write_time"),
            sum_per_sec("local_blk_read_time"),
            sum_per_sec("local_blk_write_time"),
            sum_per_sec("temp_blk_read_time"),
            sum_per_sec("temp_blk_write_time"),
            "sum(runtime) / greatest(sum(calls), 1) AS avg_runtime",
        ]

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        ):
            cols.extend(
                [
                    "sum(plantime) / greatest(sum(calls), 1) AS avg_plantime",
                    sum_per_sec("wal_records"),
                    sum_per_sec("wal_fpi"),
                    sum_per_sec("wal_bytes"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_functions"),
                    sum_per_sec("jit_generation_time"),
                    sum_per_sec("jit_inlining_count"),
                    sum_per_sec("jit_inlining_time"),
                    sum_per_sec("jit_optimization_count"),
                    sum_per_sec("jit_optimization_time"),
                    sum_per_sec("jit_emission_count"),
                    sum_per_sec("jit_emission_time"),
                ]
            )

        if self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        ):
            cols.extend(
                [
                    sum_per_sec("jit_deform_count"),
                    sum_per_sec("jit_deform_time"),
                    sum_per_sec(
                        "jit_generation_time - jit_deform_time",
                        alias="jit_expr_time",
                    ),
                ]
            )

        from_clause = query
        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            # Add system metrics from pg_stat_kcache,
            # and detailed hit ratio.
            kcache_query = kcache_getstatdata_sample(
                "query", ["datname = %(database)s", "queryid = %(query)s"]
            )

            sys_hits = (
                "greatest(sum(shared_blks_read) - sum(sub.reads), 0)"
                " * block_size"
            )
            sys_hit_ratio = (
                "{sys_hits}::numeric * 100 / ({total_blocks}"
                " * block_size)".format(
                    sys_hits=sys_hits, total_blocks=total_blocks
                )
            )
            disk_hit_ratio = (
                "sum(sub.reads) * 100 / ({total_blocks} * block_size)".format(
                    total_blocks=total_blocks
                )
            )
            total_time = "greatest(sum(runtime), 1)"
            other_time = (
                "sum(runtime) - (((sum(user_time) + sum(system_time))) * 1000)"
            )

            # Rusage can return values > real time due to sampling bias
            # aligned to kernel ticks. As such, we have to clamp values to 100%
            def total_time_percent(col, alias=None, noalias=False):
                val = "least(100, ({col} * 100) / {total_time})".format(
                    col=col, total_time=total_time
                )

                if not noalias:
                    val += " as " + alias

                return val

            cols.extend(
                [
                    sum_per_sec("reads"),
                    sum_per_sec("writes"),
                    sum_per_sec("minflts"),
                    sum_per_sec("majflts"),
                    # sum_per_sec("nswaps"),
                    # sum_per_sec("msgsnds"),
                    # sum_per_sec("msgrcvs"),
                    # sum_per_sec("nsignals"),
                    sum_per_sec("nvcsws"),
                    sum_per_sec("nivcsws"),
                    total_time_percent("sum(user_time) * 1000", "user_time"),
                    total_time_percent(
                        "sum(system_time) * 1000", "system_time"
                    ),
                    "greatest({other}, 0) AS other_time".format(
                        other=total_time_percent(other_time, noalias=True)
                    ),
                    "CASE WHEN {tb} = 0 THEN 0 ELSE {dhr} END AS disk_hit_ratio".format(
                        tb=total_blocks, dhr=disk_hit_ratio
                    ),
                    "CASE WHEN {tb} = 0 THEN 0 ELSE {shr} END AS sys_hit_ratio".format(
                        tb=total_blocks, shr=sys_hit_ratio
                    ),
                ]
            )

            from_clause = """SELECT *
                FROM (
                    {from_clause}
                ) sub2
                LEFT JOIN (
                    {kcache_query}
                ) AS kc
                    USING (ts, srvid, queryid, userid, dbid)""".format(
                from_clause=from_clause, kcache_query=kcache_query
            )
        else:
            cols.extend(
                [
                    """CASE WHEN {tb} = 0 THEN 0
                    ELSE sum(shared_blks_read)::numeric * 100 / {tb}
                END AS miss_ratio""".format(tb=total_blocks)
                ]
            )

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        CROSS JOIN {bs}
        WHERE calls != 0
        GROUP BY ts, block_size, mesure_interval
        ORDER BY ts""".format(
            cols=", ".join(cols), from_clause=from_clause, bs=block_size
        )


class QueryIndexes(ContentWidget):
    """
    Content widget showing index creation suggestion.
    """

    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/indexes"
    title = "Index Suggestions"

    def get(self, srvid, database, query):
        if not self.has_extension(srvid, "pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")

        try:
            # Check remote access first
            remote_conn = self.connect(
                srvid, database=database, remote_access=True
            )
        except Exception as e:
            raise HTTPError(
                501, "Could not connect to remote server: %s" % str(e)
            )

        extra_join = """,
        LATERAL unnest(quals) AS qual"""
        base_query = qualstat_getstatdata(
            extra_join=extra_join,
            extra_where=["queryid = " + query],
            extra_having=[
                "bool_or(eval_type = 'f')",
                "sum(execution_count) > 1000",
                "sum(occurences) > 0",
                QUALSTAT_FILTER_RATIO + " > 0.5",
            ],
        )
        optimizable = list(
            self.execute(
                base_query,
                params={
                    "server": srvid,
                    "query": query,
                    "from": "-infinity",
                    "to": "infinity",
                },
            )
        )
        optimizable = resolve_quals(remote_conn, optimizable, "quals")
        hypoplan = None
        indexes = {}
        for qual in optimizable:
            indexes[qual.where_clause] = possible_indexes(qual)
        hypo_version = self.has_extension_version(
            srvid, "hypopg", "0.0.3", database=database
        )
        if indexes and hypo_version:
            # identify indexes
            # create them
            allindexes = [
                ind
                for indcollection in indexes.values()
                for ind in indcollection
            ]
            for ind in allindexes:
                (sql, params) = ind.hypo_ddl
                if sql is not None:
                    try:
                        ind.name = self.execute(
                            sql,
                            params=params,
                            srvid=srvid,
                            database=database,
                            remote_access=True,
                        )[0]["indexname"]
                    except Error as e:
                        self.flash(
                            "Could not create hypothetical index: %s" % str(e)
                        )
            # Build the query and fetch the plans
            querystr = get_any_sample_query(
                self,
                srvid,
                database,
                query,
                self.get_argument("from"),
                self.get_argument("to"),
            )
            try:
                hypoplan = get_hypoplans(
                    remote_conn.cursor(), querystr, allindexes
                )
            except Error as e:
                # TODO: offer the possibility to fill in parameters from the UI
                self.flash(
                    "We couldn't get plans for this query, presumably "
                    "because some parameters are missing: %s" % str(e)
                )

        self.render_json(dict(indexes=indexes, hypoplan=hypoplan))


class QueryExplains(ContentWidget):
    """
    Content widget showing explain plans for various const values.
    """

    title = "Example Values"
    data_url = (
        r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/explains"
    )

    def get(self, server, database, query):
        if not self.has_extension(server, "pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")

        plans = []
        row = qualstat_get_figures(
            self,
            server,
            database,
            self.get_argument("from"),
            self.get_argument("to"),
            queries=[query],
        )
        if row is not None:
            plans = get_plans(self, server, database, row["query"], row)

        if len(plans) == 0:
            self.render_json(None)
            return

        self.render_json([plan._asdict() for plan in plans])


class WaitsQueryOverviewMetricGroup(MetricGroupDef):
    """
    Metric Group for the wait event graph on the by query page.
    """

    name = "waits_query_overview"
    xaxis = "ts"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/wait_events_sampled"
    # pg 9.6 only metrics
    count_lwlocknamed = MetricDef(
        label="Lightweight Named",
        desc="Number of named lightweight lock wait events",
    )
    count_lwlocktranche = MetricDef(
        label="Lightweight Tranche",
        desc="Number of lightweight lock tranche wait events",
    )
    # pg 10+ metrics
    count_lwlock = MetricDef(
        label="Lightweight Lock",
        desc="Number of wait events due to lightweight locks",
    )
    count_lock = MetricDef(
        label="Lock", desc="Number of wait events due to heavyweight locks"
    )
    count_bufferpin = MetricDef(
        label="Buffer pin", desc="Number of wait events due to buffer pin"
    )
    count_activity = MetricDef(
        label="Activity",
        desc="Number of wait events due to postgres"
        " internal processes activity",
    )
    count_client = MetricDef(
        label="Client", desc="Number of wait events due to client activity"
    )
    count_extension = MetricDef(
        label="Extension",
        desc="Number wait events due to third-party extensions",
    )
    count_ipc = MetricDef(
        label="IPC",
        desc="Number of wait events due to inter-process communication",
    )
    count_timeout = MetricDef(
        label="Timeout", desc="Number of wait events due to timeouts"
    )
    count_io = MetricDef(
        label="IO", desc="Number of wait events due to IO operations"
    )

    def prepare(self):
        if not self.has_extension(self.path_args[0], "pg_wait_sampling"):
            raise HTTPError(501, "pg_wait_sampling is not installed")

    @property
    def query(self):
        query = powa_getwaitdata_sample(
            "query", ["datname = %(database)s", "queryid = %(query)s"]
        )
        cols = [to_epoch("ts")]

        pg_version_num = self.get_pg_version_num(self.path_args[0])
        # if we can't connect to the remote server, assume pg10 or above
        if pg_version_num is not None and pg_version_num < 100000:
            cols += [
                wps("count_lwlocknamed", do_sum=False),
                wps("count_lwlocktranche", do_sum=False),
                wps("count_lock", do_sum=False),
                wps("count_bufferpin", do_sum=False),
            ]
        else:
            cols += [
                wps("count_lwlock", do_sum=False),
                wps("count_lock", do_sum=False),
                wps("count_bufferpin", do_sum=False),
                wps("count_activity", do_sum=False),
                wps("count_client", do_sum=False),
                wps("count_extension", do_sum=False),
                wps("count_ipc", do_sum=False),
                wps("count_timeout", do_sum=False),
                wps("count_io", do_sum=False),
            ]

        from_clause = query

        return """SELECT {cols}
        FROM (
            {from_clause}
        ) AS sub
        -- WHERE count IS NOT NULL
        ORDER BY ts""".format(cols=", ".join(cols), from_clause=from_clause)


class WaitSamplingList(MetricGroupDef):
    """
    Datasource used for the wait events grid.
    """

    name = "query_wait_events"
    xaxis = "event"
    axis_type = "category"
    data_url = (
        r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/wait_events"
    )
    counts = MetricDef(
        label="# of events", type="integer", direction="descending"
    )

    def prepare(self):
        if not self.has_extension(self.path_args[0], "pg_wait_sampling"):
            raise HTTPError(501, "pg_wait_sampling is not installed")

    @property
    def query(self):
        # Working from the waitdata detailed_db base query
        inner_query = powa_getwaitdata_detailed_db()

        cols = [
            "queryid",
            "ps.query",
            "event_type",
            "event",
            "sum(count) AS counts",
        ]

        from_clause = """(
                {inner_query}
            ) AS sub
            JOIN {{powa}}.powa_statements AS ps
                USING (srvid, queryid, dbid)""".format(inner_query=inner_query)

        return """SELECT {cols}
        FROM {from_clause}
        WHERE datname = %(database)s AND queryid = %(query)s
        GROUP BY queryid, query, event_type, event
        ORDER BY sum(count) DESC""".format(
            cols=", ".join(cols), from_clause=from_clause
        )


class QualList(MetricGroupDef):
    """
    Datasource used for the Qual table.
    """

    name = "query_quals"
    xaxis = "relname"
    axis_type = "category"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/quals"
    filter_ratio = MetricDef(
        label="Avg filter_ratio (excluding index)", type="percent"
    )
    execution_count = MetricDef(
        label="Execution count (excluding index)", type="integer"
    )

    def prepare(self):
        if not self.has_extension(self.path_args[0], "pg_qualstats"):
            raise HTTPError(501, "PG qualstats is not installed")

    @property
    def query(self):
        base = qualstat_getstatdata(extra_where=["queryid = %(query)s"])
        return base

    def post_process(self, data, server, database, query, **kwargs):
        try:
            remote_conn = self.connect(
                server, database=database, remote_access=True
            )
        except Exception as e:
            raise HTTPError(
                501, "Could not connect to remote server: %s" % str(e)
            )

        data["data"] = resolve_quals(remote_conn, data["data"])
        for qual in data["data"]:
            qual.url = self.reverse_url(
                "QualOverview", server, database, query, qual.qualid
            )
        return data


class QueryDetail(ContentWidget):
    """
    Detail widget showing summarized information for the query.
    """

    title = "Query Detail"
    data_url = r"/server/(\d+)/metrics/database/([^\/]+)/query/(-?\d+)/detail"

    def get(self, srvid, database, query):
        stmt = powa_getstatdata_detailed_db(
            srvid, ["datname = %(database)s", "queryid = %(query)s"]
        )

        from_clause = """{{powa}}.powa_statements AS ps
        LEFT JOIN (
            {stmt}
        ) AS sub USING (queryid, dbid, userid)
        CROSS JOIN {block_size}""".format(stmt=stmt, block_size=block_size)

        rblk = "(sum(shared_blks_read) * block_size)"
        wblk = "(sum(shared_blks_hit) * block_size)"
        cols = [
            "query",
            "sum(calls) AS calls",
            "sum(runtime) AS runtime",
            rblk + " AS shared_blks_read",
            wblk + " AS shared_blks_hit",
            rblk + " + " + wblk + " AS total_blks",
        ]

        stmt = """SELECT {cols}
        FROM {from_clause}
        WHERE sub.queryid = %(query)s
        AND sub.srvid = %(server)s
        GROUP BY query, block_size""".format(
            cols=", ".join(cols), from_clause=from_clause
        )

        value = self.execute(
            stmt,
            params={
                "server": srvid,
                "query": query,
                "database": database,
                "from": self.get_argument("from"),
                "to": self.get_argument("to"),
            },
        )
        if value is None or len(value) < 1:
            self.render_json(None)
            return
        self.render_json(value[0])


class QueryOverview(DashboardPage):
    """
    Dashboard page for a query.
    """

    base_url = r"/server/(\d+)/database/([^\/]+)/query/(-?\d+)/overview"
    params = ["server", "database", "query"]
    datasources = [
        QueryOverviewMetricGroup,
        WaitsQueryOverviewMetricGroup,
        QueryDetail,
        QueryExplains,
        QueryIndexes,
        WaitSamplingList,
        QualList,
        ConfigChangesQuery,
    ]
    parent = DatabaseOverview
    title = "Query Overview"
    timeline = ConfigChangesQuery

    def dashboard(self):
        # This COULD be initialized in the constructor, but tornado < 3 doesn't
        # call it
        if getattr(self, "_dashboard", None) is not None:
            return self._dashboard

        pgss18 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.8"
        )
        pgss110 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.10"
        )
        pgss111 = self.has_extension_version(
            self.path_args[0], "pg_stat_statements", "1.11"
        )

        hit_ratio_graph = Graph(
            "Hit ratio",
            metrics=[QueryOverviewMetricGroup.hit_ratio],
            renderer="bar",
            stack=True,
            color_scheme=["#73c03a", "#65b9ac", "#cb513a"],
        )

        gen_metrics = [
            QueryOverviewMetricGroup.avg_runtime,
            QueryOverviewMetricGroup.rows,
            QueryOverviewMetricGroup.calls,
        ]
        if pgss18:
            gen_metrics.extend([QueryOverviewMetricGroup.avg_plantime])

        dashes = []
        dashes.append(
            Dashboard(
                "Query detail", [[Graph("General", metrics=gen_metrics)]]
            )
        )

        if pgss18:
            # Add WALs graphs
            wals_graphs = [
                [
                    Graph(
                        "WAL activity",
                        metrics=[
                            QueryOverviewMetricGroup.wal_records,
                            QueryOverviewMetricGroup.wal_fpi,
                            QueryOverviewMetricGroup.wal_bytes,
                        ],
                    ),
                ]
            ]
            dashes.append(Dashboard("WALs", wals_graphs))

        # Add JIT graphs
        if pgss110:
            jit_tim = [
                QueryOverviewMetricGroup.jit_inlining_time,
                QueryOverviewMetricGroup.jit_optimization_time,
                QueryOverviewMetricGroup.jit_emission_time,
            ]

            if pgss111:
                jit_tim.extend(
                    [
                        QueryOverviewMetricGroup.jit_deform_time,
                        QueryOverviewMetricGroup.jit_expr_time,
                    ]
                )
            else:
                jit_tim.append(QueryOverviewMetricGroup.jit_generation_time)

            jit_cnt = [
                QueryOverviewMetricGroup.jit_functions,
                QueryOverviewMetricGroup.jit_inlining_count,
                QueryOverviewMetricGroup.jit_optimization_count,
                QueryOverviewMetricGroup.jit_emission_count,
            ]

            if pgss111:
                jit_cnt.append(QueryOverviewMetricGroup.jit_deform_count)

            jit_graphs = [
                [Graph("JIT timing", metrics=jit_tim, stack=True)],
                [Graph("JIT scheduling", metrics=jit_cnt)],
            ]

            dashes.append(Dashboard("JIT", jit_graphs))

        dashes.append(
            Dashboard(
                "PG Cache",
                [
                    [
                        Graph(
                            "Shared block (in Bps)",
                            metrics=[
                                QueryOverviewMetricGroup.shared_blks_read,
                                QueryOverviewMetricGroup.shared_blks_hit,
                                QueryOverviewMetricGroup.shared_blks_dirtied,
                                QueryOverviewMetricGroup.shared_blks_written,
                            ],
                        ),
                        Graph(
                            "Local block (in Bps)",
                            metrics=[
                                QueryOverviewMetricGroup.local_blks_read,
                                QueryOverviewMetricGroup.local_blks_hit,
                                QueryOverviewMetricGroup.local_blks_dirtied,
                                QueryOverviewMetricGroup.local_blks_written,
                            ],
                        ),
                        Graph(
                            "Temp block (in Bps)",
                            metrics=[
                                QueryOverviewMetricGroup.temp_blks_read,
                                QueryOverviewMetricGroup.temp_blks_written,
                            ],
                        ),
                    ]
                ],
            )
        )

        io_time_metrics = [
            QueryOverviewMetricGroup.shared_blk_read_time,
            QueryOverviewMetricGroup.shared_blk_write_time,
        ]

        # if we can't connect to the remote server, assume pg16 or below
        if pgss111:
            io_time_metrics.extend(
                [
                    QueryOverviewMetricGroup.local_blk_read_time,
                    QueryOverviewMetricGroup.local_blk_write_time,
                    QueryOverviewMetricGroup.temp_blk_read_time,
                    QueryOverviewMetricGroup.temp_blk_write_time,
                ]
            )
        iodash = Dashboard(
            "IO",
            [
                [
                    hit_ratio_graph,
                    Graph(
                        "Read / Write time",
                        url=self.docs_stats_url + "pg_stat_statements.html",
                        metrics=io_time_metrics,
                    ),
                ]
            ],
        )
        dashes.append(iodash)

        if self.has_extension(self.path_args[0], "pg_stat_kcache"):
            iodash.widgets.extend(
                [
                    [
                        Graph(
                            "Physical block (in Bps)",
                            url=self.docs_stats_url + "pg_stat_kcache.html",
                            metrics=[
                                QueryOverviewMetricGroup.reads,
                                QueryOverviewMetricGroup.writes,
                            ],
                        ),
                        Graph(
                            "CPU Time repartition",
                            url=self.docs_stats_url + "pg_stat_kcache.html",
                            metrics=[
                                QueryOverviewMetricGroup.user_time,
                                QueryOverviewMetricGroup.system_time,
                                QueryOverviewMetricGroup.other_time,
                            ],
                            renderer="bar",
                            stack=True,
                            color_scheme=["#73c03a", "#cb513a", "#65b9ac"],
                        ),
                    ]
                ]
            )
            hit_ratio_graph.metrics.append(
                QueryOverviewMetricGroup.sys_hit_ratio
            )
            hit_ratio_graph.metrics.append(
                QueryOverviewMetricGroup.disk_hit_ratio
            )

            sys_graphs = [
                Graph(
                    "System resources (events per sec)",
                    url=self.docs_stats_url + "pg_stat_kcache.html",
                    metrics=[
                        QueryOverviewMetricGroup.majflts,
                        QueryOverviewMetricGroup.minflts,
                        # QueryOverviewMetricGroup.nswaps,
                        # QueryOverviewMetricGroup.msgsnds,
                        # QueryOverviewMetricGroup.msgrcvs,
                        # QueryOverviewMetricGroup.nsignals,
                        QueryOverviewMetricGroup.nvcsws,
                        QueryOverviewMetricGroup.nivcsws,
                    ],
                )
            ]
            dashes.append(Dashboard("System resources", [sys_graphs]))
        else:
            hit_ratio_graph.metrics.append(QueryOverviewMetricGroup.miss_ratio)

        if self.has_extension(self.path_args[0], "pg_wait_sampling"):
            # Get the metrics depending on the pg server version
            metrics = None
            pg_version_num = self.get_pg_version_num(self.path_args[0])
            # if we can't connect to the remote server, assume pg10 or above
            if pg_version_num is not None and pg_version_num < 100000:
                metrics = [
                    WaitsQueryOverviewMetricGroup.count_lwlocknamed,
                    WaitsQueryOverviewMetricGroup.count_lwlocktranche,
                    WaitsQueryOverviewMetricGroup.count_lock,
                    WaitsQueryOverviewMetricGroup.count_bufferpin,
                ]
            else:
                metrics = [
                    WaitsQueryOverviewMetricGroup.count_lwlock,
                    WaitsQueryOverviewMetricGroup.count_lock,
                    WaitsQueryOverviewMetricGroup.count_bufferpin,
                    WaitsQueryOverviewMetricGroup.count_activity,
                    WaitsQueryOverviewMetricGroup.count_client,
                    WaitsQueryOverviewMetricGroup.count_extension,
                    WaitsQueryOverviewMetricGroup.count_ipc,
                    WaitsQueryOverviewMetricGroup.count_timeout,
                    WaitsQueryOverviewMetricGroup.count_io,
                ]
            dashes.append(
                Dashboard(
                    "Wait Events",
                    [
                        [
                            Graph(
                                "Wait Events (per second)",
                                url=self.docs_stats_url
                                + "pg_wait_sampling.html",
                                metrics=metrics,
                            ),
                            Grid(
                                "Wait events summary",
                                url=self.docs_stats_url
                                + "pg_wait_sampling.html",
                                columns=[
                                    {
                                        "name": "event_type",
                                        "label": "Event Type",
                                    },
                                    {
                                        "name": "event",
                                        "label": "Event",
                                    },
                                ],
                                metrics=WaitSamplingList.all(),
                            ),
                        ]
                    ],
                )
            )

        if self.has_extension(self.path_args[0], "pg_qualstats"):
            dashes.append(
                Dashboard(
                    "Predicates",
                    [
                        [
                            Grid(
                                "Predicates used by this query",
                                columns=[
                                    {
                                        "name": "where_clause",
                                        "label": "Predicate",
                                        "type": "query",
                                        "max_length": 60,
                                        "url_attr": "url",
                                    }
                                ],
                                metrics=QualList.all(),
                            )
                        ],
                        [QueryIndexes],
                        [QueryExplains],
                    ],
                )
            )
        self._dashboard = Dashboard(
            "Query %(query)s on database %(database)s",
            [
                [QueryDetail],
                [
                    TabContainer(
                        "Query %(query)s on database %(database)s", dashes
                    )
                ],
            ],
        )
        return self._dashboard
